from __future__ import annotations

import asyncio
import os
import sys
from pathlib import Path, PureWindowsPath
from random import randint
from typing import Any, Optional

from aiohttp import ClientConnectorError
from chik_rs import PrivateKey

from chik.cmds.cmds_util import get_any_service_client
from chik.cmds.start_funcs import async_start
from chik.consensus.coinbase import create_puzzlehash_for_pk
from chik.server.outbound_message import NodeType
from chik.simulator.simulator_full_node_rpc_client import SimulatorFullNodeRpcClient
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_record import CoinRecord
from chik.util.bech32m import decode_puzzle_hash, encode_puzzle_hash
from chik.util.config import load_config, save_config, set_peer_info
from chik.util.errors import KeychainFingerprintExists
from chik.util.ints import uint32
from chik.util.keychain import Keychain, bytes_to_mnemonic
from chik.wallet.derive_keys import (
    master_sk_to_farmer_sk,
    master_sk_to_pool_sk,
    master_sk_to_wallet_sk,
    master_sk_to_wallet_sk_unhardened,
)


def get_ph_from_fingerprint(fingerprint: int, key_id: int = 1) -> bytes32:
    priv_key_and_entropy = Keychain().get_private_key_by_fingerprint(fingerprint)
    if priv_key_and_entropy is None:
        raise Exception("Fingerprint not found")
    private_key = priv_key_and_entropy[0]
    sk_for_wallet_id: PrivateKey = master_sk_to_wallet_sk(private_key, uint32(key_id))
    puzzle_hash: bytes32 = create_puzzlehash_for_pk(sk_for_wallet_id.get_g1())
    return puzzle_hash


def create_chik_directory(
    chik_root: Path,
    fingerprint: int,
    farming_address: Optional[str],
    plot_directory: Optional[str],
    auto_farm: Optional[bool],
    docker_mode: bool,
) -> dict[str, Any]:
    """
    This function creates a new chik directory and returns a heavily modified config,
    suitable for use in the simulator.
    """
    from chik.cmds.init_funcs import chik_init

    if not chik_root.is_dir() or not Path(chik_root / "config" / "config.yaml").exists():
        # create chik directories & load config
        chik_init(chik_root, testnet=True, fix_ssl_permissions=True)
        config: dict[str, Any] = load_config(chik_root, "config.yaml")
        # apply standard block-tools config.
        config["full_node"]["send_uncompact_interval"] = 0
        config["full_node"]["target_uncompact_proofs"] = 30
        config["full_node"]["peer_connect_interval"] = 50
        config["full_node"]["sanitize_weight_proof_only"] = False
        config["logging"]["log_level"] = "INFO"  # extra logs for easier development
        # make sure we don't try to connect to other nodes.
        config["full_node"]["introducer_peer"] = None
        config["wallet"]["introducer_peer"] = None
        config["full_node"]["dns_servers"] = []
        config["wallet"]["dns_servers"] = []
        # create custom testnet (simulator0)
        config["network_overrides"]["constants"]["simulator0"] = config["network_overrides"]["constants"][
            "testnet0"
        ].copy()
        config["network_overrides"]["config"]["simulator0"] = config["network_overrides"]["config"]["testnet0"].copy()
        sim_genesis = "eb8c4d20b322be8d9fddbf9412016bdffe9a2901d7edb0e364e94266d0e095f7"
        config["network_overrides"]["constants"]["simulator0"]["GENESIS_CHALLENGE"] = sim_genesis
        # tell services to use simulator0
        config["selected_network"] = "simulator0"
        config["wallet"]["selected_network"] = "simulator0"
        config["full_node"]["selected_network"] = "simulator0"
        config["seeder"]["selected_network"] = "simulator0"
        config["harvester"]["selected_network"] = "simulator0"
        config["pool"]["selected_network"] = "simulator0"
        config["farmer"]["selected_network"] = "simulator0"
        config["timelord"]["selected_network"] = "simulator0"
        config["ui"]["selected_network"] = "simulator0"
        config["introducer"]["selected_network"] = "simulator0"
        config["data_layer"]["selected_network"] = "simulator0"
        if not docker_mode:  # We want predictable ports for our docker image.
            # set ports and networks, we don't want to cause a port conflict.
            port_offset = randint(1, 20000)
            config["daemon_port"] -= port_offset
            config["network_overrides"]["config"]["simulator0"]["default_full_node_port"] = 39678 + port_offset
            # wallet
            config["wallet"]["rpc_port"] += port_offset
            # full node
            config["full_node"]["port"] -= port_offset
            config["full_node"]["rpc_port"] += port_offset
            # connect wallet to full node
            set_peer_info(config["wallet"], peer_type=NodeType.FULL_NODE, peer_port=config["full_node"]["port"])
            # ui
            config["ui"]["daemon_port"] = config["daemon_port"]
        else:
            config["self_hostname"] = "0.0.0.0"  # Bind to all interfaces.
            config["logging"]["log_stdout"] = True  # Log to console.
    else:
        config = load_config(chik_root, "config.yaml")
    # simulator overrides
    sim_config = config["simulator"]
    sim_config["key_fingerprint"] = fingerprint
    if farming_address is None:
        prefix = config["network_overrides"]["config"]["simulator0"]["address_prefix"]
        farming_address = encode_puzzle_hash(get_ph_from_fingerprint(fingerprint), prefix)
    sim_config["farming_address"] = farming_address
    if plot_directory is not None:
        sim_config["plot_directory"] = plot_directory
    # Temporary change to fix win / linux differences.
    sim_config["plot_directory"] = str(Path(sim_config["plot_directory"]))
    if "//" in sim_config["plot_directory"] and os.name != "nt":
        # if we're on linux, we need to convert to a linux path.
        sim_config["plot_directory"] = str(PureWindowsPath(sim_config["plot_directory"]).as_posix())
    sim_config["auto_farm"] = auto_farm if auto_farm is not None else True
    farming_ph = decode_puzzle_hash(farming_address)
    # modify genesis block to give the user the reward
    simulator_consts = config["network_overrides"]["constants"]["simulator0"]
    simulator_consts["GENESIS_PRE_FARM_FARMER_PUZZLE_HASH"] = farming_ph.hex()
    simulator_consts["GENESIS_PRE_FARM_POOL_PUZZLE_HASH"] = farming_ph.hex()
    # get fork heights then write back to config
    if "HARD_FORK_HEIGHT" not in sim_config:  # this meh code is done so that we also write to the config file.
        sim_config["HARD_FORK_HEIGHT"] = 0
    if "SOFT_FORK6_HEIGHT" not in sim_config:
        sim_config["SOFT_FORK6_HEIGHT"] = 0
    simulator_consts["HARD_FORK_HEIGHT"] = sim_config["HARD_FORK_HEIGHT"]
    simulator_consts["SOFT_FORK6_HEIGHT"] = sim_config["SOFT_FORK6_HEIGHT"]

    # save config and return the config
    save_config(chik_root, "config.yaml", config)
    return config


def display_key_info(fingerprint: int, prefix: str) -> None:
    """
    Display key info for a given fingerprint, similar to the output of `chik keys show`.
    """
    print(f"Using fingerprint {fingerprint}")
    private_key_and_seed = Keychain().get_private_key_by_fingerprint(fingerprint)
    if private_key_and_seed is None:
        print(f"Fingerprint {fingerprint} not found")
        return
    sk, seed = private_key_and_seed
    print("\nFingerprint:", sk.get_g1().get_fingerprint())
    print("Master public key (m):", sk.get_g1())
    print("Farmer public key (m/12381/8444/0/0):", master_sk_to_farmer_sk(sk).get_g1())
    print("Pool public key (m/12381/8444/1/0):", master_sk_to_pool_sk(sk).get_g1())
    first_wallet_sk: PrivateKey = master_sk_to_wallet_sk_unhardened(sk, uint32(0))
    wallet_address: str = encode_puzzle_hash(create_puzzlehash_for_pk(first_wallet_sk.get_g1()), prefix)
    print(f"First wallet address: {wallet_address}")
    assert seed is not None
    print("Master private key (m):", bytes(sk).hex())
    print("First wallet secret key (m/12381/8444/2/0):", master_sk_to_wallet_sk(sk, uint32(0)))
    mnemonic = bytes_to_mnemonic(seed)
    print("  Mnemonic seed (24 secret words):")
    print(f"{mnemonic} \n")


def generate_and_return_fingerprint(mnemonic: Optional[str] = None) -> int:
    """
    Generate and add new PrivateKey and return its fingerprint.
    """
    from chik.util.keychain import generate_mnemonic

    if mnemonic is None:
        print("Generating private key")
        mnemonic = generate_mnemonic()
    try:
        sk = Keychain().add_key(mnemonic, None)
        fingerprint: int = sk.get_g1().get_fingerprint()
    except KeychainFingerprintExists as e:
        fingerprint = e.fingerprint
        print(f"Fingerprint: {fingerprint} for provided private key already exists.")
        return fingerprint
    print(f"Added private key with public key fingerprint {fingerprint}")
    return fingerprint


def select_fingerprint(
    fingerprint: Optional[int] = None, mnemonic_string: Optional[str] = None, auto_generate_key: bool = False
) -> Optional[int]:
    """
    Either select an existing fingerprint or create one and return it.
    """
    if mnemonic_string:
        fingerprint = generate_and_return_fingerprint(mnemonic_string)
    fingerprints: list[int] = [pk.get_fingerprint() for pk in Keychain().get_all_public_keys()]
    if fingerprint is not None and fingerprint in fingerprints:
        return fingerprint
    elif fingerprint is not None and fingerprint not in fingerprints:
        print(f"Invalid Fingerprint. Fingerprint {fingerprint} was not found.")
        return None
    if auto_generate_key and len(fingerprints) == 1:
        return fingerprints[0]
    if len(fingerprints) == 0:
        if not auto_generate_key:
            if (
                input("No keys in keychain. Press 'q' to quit, or press any other key to generate a new key.").lower()
                == "q"
            ):
                return None
        # generate private key and add to wallet
        fingerprint = generate_and_return_fingerprint()
    else:
        print("Fingerprints:")
        print(
            "If you already used one of these keys, select that fingerprint to skip the plotting process."
            " Otherwise, select any key below."
        )
        for i, fp in enumerate(fingerprints):
            row: str = f"{i + 1}) "
            row += f"{fp}"
            print(row)
        val = None
        prompt: str = f"Choose a simulator key [1-{len(fingerprints)}] ('q' to quit, or 'g' to generate a new key): "
        while val is None:
            val = input(prompt)
            if val == "q":
                return None
            elif val == "g":
                fingerprint = generate_and_return_fingerprint()
                break
            elif not val.isdigit():
                val = None
            else:
                index = int(val) - 1
                if index < 0 or index >= len(fingerprints):
                    print("Invalid value")
                    val = None
                    continue
                else:
                    fingerprint = fingerprints[index]
        assert fingerprint is not None
    return fingerprint


async def generate_plots(config: dict[str, Any], root_path: Path, fingerprint: int, bitfield: bool) -> None:
    """
    Pre-Generate plots for the new simulator instance.
    """

    from chik.simulator.block_tools import BlockTools, test_constants
    from chik.simulator.start_simulator import PLOT_SIZE, PLOTS

    farming_puzzle_hash = decode_puzzle_hash(config["simulator"]["farming_address"])
    os.environ["CHIK_ROOT"] = str(root_path)  # change env variable, to make it match what the daemon would set it to

    # create block tools and use local keychain
    bt = BlockTools(
        test_constants,
        root_path,
        automated_testing=False,
        plot_dir=config["simulator"].get("plot_directory", "plots"),
        keychain=Keychain(),
    )
    await bt.setup_keys(fingerprint=fingerprint, reward_ph=farming_puzzle_hash)
    existing_plots = await bt.setup_plots(
        num_og_plots=PLOTS, num_pool_plots=0, num_non_keychain_plots=0, plot_size=PLOT_SIZE, bitfield=bitfield
    )
    print(f"{'New plots generated.' if existing_plots else 'Using Existing Plots'}\n")


async def get_current_height(root_path: Path) -> int:
    async with get_any_service_client(SimulatorFullNodeRpcClient, root_path=root_path, consume_errors=False) as (
        node_client,
        _,
    ):
        num_blocks = len(await node_client.get_all_blocks())
    return num_blocks


async def async_config_wizard(
    root_path: Path,
    fingerprint: Optional[int],
    farming_address: Optional[str],
    plot_directory: Optional[str],
    mnemonic_string: Optional[str],
    auto_farm: Optional[bool],
    docker_mode: bool,
    bitfield: bool,
) -> None:
    # either return passed through fingerprint or get one
    fingerprint = select_fingerprint(fingerprint, mnemonic_string, docker_mode)
    if fingerprint is None:
        # user cancelled wizard
        return
    # create chik directory & get config.
    print("Creating chik directory & config...")
    config = create_chik_directory(root_path, fingerprint, farming_address, plot_directory, auto_farm, docker_mode)
    # Pre-generate plots by running block_tools init functions.
    print("Please Wait, Generating plots...")
    print("This may take up to a minute if you are on a slow machine")

    await generate_plots(config, root_path, fingerprint, bitfield)
    # final messages
    final_farming_address = config["simulator"]["farming_address"]
    print(f"\nFarming & Prefarm reward address: {final_farming_address}\n")
    print("Configuration Wizard Complete.")
    print("Starting Simulator now...\n\n")

    sys.argv[0] = str(Path(sys.executable).parent / "chik")  # fix path for tests
    await async_start(root_path, config, ("simulator",), restart=False, skip_keyring=False)

    # now we make sure the simulator has a genesis block
    print("Please wait, generating genesis block.")
    while True:
        try:
            num_blocks: int = await get_current_height(root_path)
        except ClientConnectorError:
            await asyncio.sleep(0.25)
        else:
            if num_blocks == 0:
                await farm_blocks(None, root_path, 1, True, final_farming_address)
                print("Genesis block generated, exiting.")
            else:
                print("Genesis block already exists, exiting.")
            break
    print(f"\nMake sure your CHIK_ROOT Environment Variable is set to: {root_path}")


def print_coin_record(
    name: str,
    address_prefix: str,
    coin_record: CoinRecord,
) -> None:
    from datetime import datetime

    coin_address = encode_puzzle_hash(coin_record.coin.puzzle_hash, address_prefix)
    print(f"Coin 0x{coin_record.name.hex()}")
    print(f"Wallet Address: {coin_address}")
    print(f"Confirmed at block: {coin_record.confirmed_block_index}")
    print(f"Spent: {f'at Block {coin_record.spent_block_index}' if coin_record.spent else 'No'}")
    print(f"Coin Amount: {coin_record.coin.amount} {name}")
    print(f"Parent Coin ID: 0x{coin_record.coin.parent_coin_info.hex()}")
    print(f"Created at: {datetime.fromtimestamp(float(coin_record.timestamp)).strftime('%Y-%m-%d %H:%M:%S')}\n")


async def print_coin_records(
    config: dict[str, Any],
    node_client: SimulatorFullNodeRpcClient,
    include_reward_coins: bool,
    include_spent: bool = False,
) -> None:
    import sys

    coin_records: list[CoinRecord] = await node_client.get_all_coins(include_spent)
    coin_records = [coin_record for coin_record in coin_records if not coin_record.coinbase or include_reward_coins]
    address_prefix = config["network_overrides"]["config"][config["selected_network"]]["address_prefix"]
    name = "mojo"
    paginate = False  # I might change this later.
    if len(coin_records) != 0:
        print("All Coins: ")
        if paginate is True:
            paginate = sys.stdout.isatty()
        num_per_screen = 5 if paginate else len(coin_records)
        # ripped from cmds/wallet_funcs.
        for i in range(0, len(coin_records), num_per_screen):
            for j in range(0, num_per_screen):
                if i + j >= len(coin_records):
                    break
                print_coin_record(
                    coin_record=coin_records[i + j],
                    name=name,
                    address_prefix=address_prefix,
                )
            if i + num_per_screen <= len(coin_records) and paginate:
                print("Press q to quit, or c to continue")
                while True:
                    entered_key = sys.stdin.read(1)
                    if entered_key == "q":
                        return None
                    elif entered_key == "c":
                        break


async def print_wallets(config: dict[str, Any], node_client: SimulatorFullNodeRpcClient) -> None:
    ph_and_amount = await node_client.get_all_puzzle_hashes()
    address_prefix = config["network_overrides"]["config"][config["selected_network"]]["address_prefix"]
    name = "mojo"
    for puzzle_hash, (amount, num_tx) in ph_and_amount.items():
        address = encode_puzzle_hash(puzzle_hash, address_prefix)
        print(f"Address: {address} has a balance of: {amount} {name}, with a total of: {num_tx} transactions.\n")


async def print_status(
    rpc_port: Optional[int],
    root_path: Path,
    fingerprint: Optional[int],
    show_key: bool,
    show_coins: bool,
    include_reward_coins: bool,
    show_addresses: bool,
) -> None:
    """
    This command allows users to easily get the status of the simulator
    and information about the state of and the coins in the simulated blockchain.
    """
    from chik.cmds.show_funcs import print_blockchain_state
    from chik.cmds.units import units

    async with get_any_service_client(SimulatorFullNodeRpcClient, root_path, rpc_port) as (node_client, config):
        # Display keychain info
        if show_key:
            if fingerprint is None:
                fingerprint = config["simulator"]["key_fingerprint"]
            if fingerprint is not None:
                display_key_info(
                    fingerprint, config["network_overrides"]["config"][config["selected_network"]]["address_prefix"]
                )
            else:
                print(
                    "No fingerprint in config, either rerun 'cdv sim create' "
                    "or use --fingerprint to specify one, skipping key information."
                )
        # chain status ( basically chik show -s)
        await print_blockchain_state(node_client, config)
        print("")
        # farming information
        target_ph: bytes32 = await node_client.get_farming_ph()
        farming_coin_records = await node_client.get_coin_records_by_puzzle_hash(target_ph, False)
        prefix = config["network_overrides"]["config"][config["selected_network"]]["address_prefix"]
        print(
            f"Current Farming address: {encode_puzzle_hash(target_ph, prefix)}, "
            f"with a balance of: "
            f"{sum(coin_records.coin.amount for coin_records in farming_coin_records) / units['chik']} TXCK."
        )
        if show_addresses:
            print("All Addresses: ")
            await print_wallets(config, node_client)
        if show_coins:
            await print_coin_records(config, node_client, include_reward_coins)


async def revert_block_height(
    rpc_port: Optional[int],
    root_path: Path,
    num_blocks: int,
    num_new_blocks: int,
    reset_chain_to_genesis: bool,
    use_revert_blocks: bool,
) -> None:
    """
    This function allows users to easily revert the chain to a previous state or perform a reorg.
    """
    async with get_any_service_client(SimulatorFullNodeRpcClient, root_path, rpc_port) as (node_client, _):
        if use_revert_blocks:
            if num_new_blocks != 1:
                print(f"Ignoring num_new_blocks: {num_new_blocks}, because we are not performing a reorg.")
            # in this case num_blocks is the number of blocks to delete
            new_height: int = await node_client.revert_blocks(num_blocks, reset_chain_to_genesis)
            print(
                f"All transactions in Block: {new_height + num_blocks} and above were successfully deleted, "
                "you should now delete & restart all wallets."
            )
        else:
            # However, in this case num_blocks is the fork height.
            new_height = await node_client.reorg_blocks(num_blocks, num_new_blocks, use_revert_blocks)
            old_height = new_height - num_new_blocks
            print(f"All transactions in Block: {old_height - num_blocks} and above were successfully reverted.")
        print(f"Block Height is now: {new_height}")


async def farm_blocks(
    rpc_port: Optional[int],
    root_path: Path,
    num_blocks: int,
    transaction_blocks: bool,
    target_address: str,
) -> None:
    """
    This function is used to generate new blocks.
    """
    async with get_any_service_client(SimulatorFullNodeRpcClient, root_path, rpc_port) as (node_client, config):
        if target_address == "":
            target_address = config["simulator"]["farming_address"]
        if target_address is None:
            print(
                "No target address in config, falling back to the temporary address currently in use. "
                "You can use 'cdv sim create' or use --target-address to specify a different address."
            )
            target_ph: bytes32 = await node_client.get_farming_ph()
        else:
            target_ph = decode_puzzle_hash(target_address)
        await node_client.farm_block(target_ph, num_blocks, transaction_blocks)
        print(f"Farmed {num_blocks}{' Transaction' if transaction_blocks else ''} blocks")
        block_height = (await node_client.get_blockchain_state())["peak"].height
        print(f"Block Height is now: {block_height}")


async def set_auto_farm(rpc_port: Optional[int], root_path: Path, set_autofarm: bool) -> None:
    """
    This function can be used to enable or disable Auto Farming.
    """
    async with get_any_service_client(SimulatorFullNodeRpcClient, root_path, rpc_port) as (node_client, _):
        current = await node_client.get_auto_farming()
        if current == set_autofarm:
            print(f"Auto farming is already {'on' if set_autofarm else 'off'}")
            return
        result = await node_client.set_auto_farming(set_autofarm)
        print(f"Auto farming is now {'on' if result else 'off'}")



# Package: utils

from __future__ import annotations

import logging
import re
import time
from re import Pattern


class TimedDuplicateFilter(logging.Filter):
    last_log_time: float
    regex: Pattern[str]
    min_time_wait_secs: int

    def __init__(self, regex_str: str, min_time_wait_secs: int, name: str = ""):
        super().__init__(name)
        self.last_log_time = time.monotonic()
        self.regex = re.compile(regex_str)
        self.min_time_wait_secs = min_time_wait_secs

    def filter(self, record: logging.LogRecord) -> bool:
        _ = super().filter(record)
        if not _:
            return False

        msg = record.getMessage()

        if self.regex.match(msg):
            now = time.monotonic()
            if now - self.last_log_time > self.min_time_wait_secs:
                self.last_log_time = now
                return True
            return False
        else:
            return True



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint8, uint32, uint64

# This is a modified version of the TestBlockRecord from test_mempool_manager.py


@dataclass(frozen=True)
class TestBlockRecord:
    """
    This is a subset of BlockRecord that the cli tests need
    """

    header_hash: bytes32
    height: uint32
    timestamp: Optional[uint64]
    prev_transaction_block_height: uint32
    prev_transaction_block_hash: Optional[bytes32]
    prev_hash: Optional[bytes32]
    weight: uint64 = uint64(10000)
    fees: uint64 = uint64(5000)
    farmer_puzzle_hash: bytes32 = bytes32([1] * 32)
    pool_puzzle_hash: bytes32 = bytes32([2] * 32)
    sub_slot_iters: uint64 = uint64(1024)
    total_iters: uint64 = uint64(12081)
    deficit: uint8 = uint8(0)

    @property
    def is_transaction_block(self) -> bool:
        return self.timestamp is not None


def height_hash(height: int) -> bytes32:
    return bytes32(height.to_bytes(32, byteorder="big"))


def hash_to_height(int_bytes: bytes32) -> int:
    return int.from_bytes(int_bytes, byteorder="big")


def create_test_block_record(
    *, height: uint32 = uint32(11), timestamp: uint64 = uint64(10040), header_hash: Optional[bytes32] = None
) -> TestBlockRecord:
    if header_hash is None:
        header_hash = height_hash(height)
    else:
        height = uint32(hash_to_height(header_hash))  # so the heights make sense
    return TestBlockRecord(
        header_hash=header_hash,
        height=height,
        timestamp=timestamp,
        prev_transaction_block_height=uint32(height - 1),
        prev_transaction_block_hash=height_hash(height - 1),
        prev_hash=height_hash(height - 1),
    )



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Callable, SupportsBytes

import pytest
from chik_rs import G1Element, G2Element

from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend, make_spend
from chik.types.spend_bundle import SpendBundle
from chik.util.ints import uint64
from chik.wallet.cat_wallet.cat_utils import CAT_MOD, construct_cat_puzzle
from chik.wallet.puzzles.p2_delegated_puzzle_or_hidden_puzzle import puzzle_for_pk
from chik.wallet.trading.offer import OFFER_MOD
from chik.wallet.util.puzzle_compression import (
    LATEST_VERSION,
    OFFER_MOD_OLD,
    compress_object_with_puzzles,
    decompress_object_with_puzzles,
    lowest_best_version,
)

ONE_32 = bytes32([17] * 32)
COIN = Coin(bytes32.zeros, bytes32.zeros, uint64(0))
SOLUTION = Program.to([])


log = logging.getLogger(__name__)


@dataclass
class CompressionReporter:
    record_property: Callable

    def __call__(self, name: str, original: SupportsBytes, compressed: SupportsBytes):
        ratio = len(bytes(compressed)) / len(bytes(original))
        self.record_property(name="compression_ratio", value=ratio)
        log.warning(f"{name} compression ratio: {ratio}")


@pytest.fixture(name="report_compression")
def report_compression_fixture(record_property):
    return CompressionReporter(record_property=record_property)


def test_standard_puzzle(report_compression):
    coin_spend = make_spend(
        COIN,
        puzzle_for_pk(G1Element()),
        SOLUTION,
    )
    compressed = compress_object_with_puzzles(bytes(coin_spend), LATEST_VERSION)
    assert len(bytes(coin_spend)) > len(compressed)
    assert coin_spend == CoinSpend.from_bytes(decompress_object_with_puzzles(compressed))

    report_compression(name="standard puzzle", original=coin_spend, compressed=compressed)


def test_decompress_limit():
    buffer = bytearray(10 * 1024 * 1024)
    compressed = compress_object_with_puzzles(buffer, LATEST_VERSION)
    print(len(compressed))
    decompressed = decompress_object_with_puzzles(compressed)
    print(len(decompressed))
    assert len(decompressed) <= 6 * 1024 * 1024


def test_cat_puzzle(report_compression):
    coin_spend = make_spend(
        COIN,
        construct_cat_puzzle(CAT_MOD, Program.to([]).get_tree_hash(), Program.to(1)),
        SOLUTION,
    )
    compressed = compress_object_with_puzzles(bytes(coin_spend), LATEST_VERSION)
    assert len(bytes(coin_spend)) > len(compressed)
    assert coin_spend == CoinSpend.from_bytes(decompress_object_with_puzzles(compressed))

    report_compression(name="CAT puzzle", original=coin_spend, compressed=compressed)


def test_offer_puzzle(report_compression):
    coin_spend = make_spend(
        COIN,
        OFFER_MOD,
        SOLUTION,
    )
    compressed = compress_object_with_puzzles(bytes(coin_spend), LATEST_VERSION)
    assert len(bytes(coin_spend)) > len(compressed)
    assert coin_spend == CoinSpend.from_bytes(decompress_object_with_puzzles(compressed))

    report_compression(name="offer puzzle", original=coin_spend, compressed=compressed)


def test_nesting_puzzles(report_compression):
    coin_spend = make_spend(
        COIN,
        construct_cat_puzzle(CAT_MOD, Program.to([]).get_tree_hash(), puzzle_for_pk(G1Element())),
        SOLUTION,
    )
    compressed = compress_object_with_puzzles(bytes(coin_spend), LATEST_VERSION)
    assert len(bytes(coin_spend)) > len(compressed)
    assert coin_spend == CoinSpend.from_bytes(decompress_object_with_puzzles(compressed))

    report_compression(name="nesting puzzle", original=coin_spend, compressed=compressed)


def test_unknown_wrapper(report_compression):
    unknown = Program.to([2, 2, []])  # (a 2 ())
    coin_spend = make_spend(
        COIN,
        unknown.curry(puzzle_for_pk(G1Element())),
        SOLUTION,
    )
    compressed = compress_object_with_puzzles(bytes(coin_spend), LATEST_VERSION)
    assert len(bytes(coin_spend)) > len(compressed)
    assert coin_spend == CoinSpend.from_bytes(decompress_object_with_puzzles(compressed))

    report_compression(name="unknown wrapper", original=coin_spend, compressed=compressed)


def test_lowest_best_version():
    assert lowest_best_version([bytes(CAT_MOD)]) == 4
    assert lowest_best_version([bytes(OFFER_MOD_OLD)]) == 2
    assert lowest_best_version([bytes(OFFER_MOD)]) == 5


def test_version_override():
    coin_spend = make_spend(
        COIN,
        OFFER_MOD,
        SOLUTION,
    )
    spend_bundle = SpendBundle([coin_spend], G2Element())
    compressed = compress_object_with_puzzles(bytes(spend_bundle), LATEST_VERSION)
    compressed_earlier = compress_object_with_puzzles(bytes(spend_bundle), 1)
    assert len(bytes(spend_bundle)) > len(bytes(compressed))
    assert spend_bundle == SpendBundle.from_bytes(decompress_object_with_puzzles(compressed))
    assert spend_bundle == SpendBundle.from_bytes(decompress_object_with_puzzles(compressed_earlier))
    assert len(bytes(compressed_earlier)) > len(bytes(compressed))



from __future__ import annotations

import copy
import dataclasses
from collections.abc import Awaitable
from typing import Callable, Optional

from chik_rs import fast_forward_singleton, get_conditions_from_spendbundle

from chik.consensus.condition_costs import ConditionCost
from chik.consensus.constants import ConsensusConstants
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.serialized_program import SerializedProgram
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend
from chik.types.internal_mempool_item import InternalMempoolItem
from chik.types.mempool_item import BundleCoinSpend
from chik.types.spend_bundle import SpendBundle
from chik.util.errors import Err
from chik.util.ints import uint32, uint64


@dataclasses.dataclass(frozen=True)
class EligibilityAndAdditions:
    is_eligible_for_dedup: bool
    spend_additions: list[Coin]
    # This is the spend puzzle hash. It's set to `None` if the spend is not
    # eligible for fast forward. When the spend is eligible, we use its puzzle
    # hash to check if the singleton has an unspent coin or not.
    ff_puzzle_hash: Optional[bytes32] = None


def run_for_cost(
    puzzle_reveal: SerializedProgram, solution: SerializedProgram, additions_count: int, max_cost: int
) -> uint64:
    create_coins_cost = additions_count * ConditionCost.CREATE_COIN.value
    klvm_cost, _ = puzzle_reveal.run_mempool_with_cost(max_cost, solution)
    saved_cost = uint64(klvm_cost + create_coins_cost)
    return saved_cost


@dataclasses.dataclass(frozen=True)
class DedupCoinSpend:
    solution: SerializedProgram
    cost: Optional[uint64]


@dataclasses.dataclass(frozen=True)
class UnspentLineageInfo:
    coin_id: bytes32
    coin_amount: uint64
    parent_id: bytes32
    parent_amount: uint64
    parent_parent_id: bytes32


def set_next_singleton_version(
    current_singleton: Coin, singleton_additions: list[Coin], fast_forward_spends: dict[bytes32, UnspentLineageInfo]
) -> None:
    """
    Finds the next version of the singleton among its additions and updates the
    fast forward spends, currently chained together, accordingly

    Args:
        current_singleton: the current iteration of the singleton
        singleton_additions: the additions of the current singleton
        fast_forward_spends: in-out parameter of the spends currently chained together

    Raises:
        ValueError if none of the additions are considered to be the singleton's
        next iteration
    """
    singleton_child = next(
        (addition for addition in singleton_additions if addition.puzzle_hash == current_singleton.puzzle_hash), None
    )
    if singleton_child is None:
        raise ValueError("Could not find fast forward child singleton.")
    # Keep track of this in order to chain the next ff
    fast_forward_spends[current_singleton.puzzle_hash] = UnspentLineageInfo(
        coin_id=singleton_child.name(),
        coin_amount=singleton_child.amount,
        parent_id=singleton_child.parent_coin_info,
        parent_amount=current_singleton.amount,
        parent_parent_id=current_singleton.parent_coin_info,
    )


def perform_the_fast_forward(
    unspent_lineage_info: UnspentLineageInfo,
    spend_data: BundleCoinSpend,
    fast_forward_spends: dict[bytes32, UnspentLineageInfo],
) -> tuple[CoinSpend, list[Coin]]:
    """
    Performs a singleton fast forward, including the updating of all previous
    additions to point to the most recent version, and updates the fast forward
    spends, currently chained together, accordingly

    Args:
        unspent_lineage_info: the singleton's most recent lineage information
        spend_data: the current spend's data
        fast_forward_spends: in-out parameter of the spends currently chained together

    Returns:
        CoinSpend: the new coin spend after performing the fast forward
        list[Coin]: the updated additions that point to the new coin to spend

    Raises:
        ValueError if none of the additions are considered to be the singleton's
        next iteration
    """
    new_coin = Coin(
        unspent_lineage_info.parent_id, spend_data.coin_spend.coin.puzzle_hash, unspent_lineage_info.coin_amount
    )
    new_parent = Coin(
        unspent_lineage_info.parent_parent_id,
        spend_data.coin_spend.coin.puzzle_hash,
        unspent_lineage_info.parent_amount,
    )
    # These hold because puzzle hash is not expected to change
    assert new_coin.name() == unspent_lineage_info.coin_id
    assert new_parent.name() == unspent_lineage_info.parent_id
    new_solution = SerializedProgram.from_bytes(
        fast_forward_singleton(spend=spend_data.coin_spend, new_coin=new_coin, new_parent=new_parent)
    )
    singleton_child = None
    patched_additions = []
    for addition in spend_data.additions:
        patched_addition = Coin(unspent_lineage_info.coin_id, addition.puzzle_hash, addition.amount)
        patched_additions.append(patched_addition)
        if addition.puzzle_hash == spend_data.coin_spend.coin.puzzle_hash:
            # We found the next version of this singleton
            singleton_child = patched_addition
    if singleton_child is None:
        raise ValueError("Could not find fast forward child singleton.")
    new_coin_spend = CoinSpend(new_coin, spend_data.coin_spend.puzzle_reveal, new_solution)
    # Keep track of this in order to chain the next ff
    fast_forward_spends[spend_data.coin_spend.coin.puzzle_hash] = UnspentLineageInfo(
        coin_id=singleton_child.name(),
        coin_amount=singleton_child.amount,
        parent_id=singleton_child.parent_coin_info,
        parent_amount=unspent_lineage_info.coin_amount,
        parent_parent_id=unspent_lineage_info.parent_id,
    )
    return new_coin_spend, patched_additions


@dataclasses.dataclass(frozen=True)
class EligibleCoinSpends:
    deduplication_spends: dict[bytes32, DedupCoinSpend] = dataclasses.field(default_factory=dict)
    fast_forward_spends: dict[bytes32, UnspentLineageInfo] = dataclasses.field(default_factory=dict)

    def get_deduplication_info(
        self, *, bundle_coin_spends: dict[bytes32, BundleCoinSpend], max_cost: int
    ) -> tuple[list[CoinSpend], uint64, list[Coin]]:
        """
        Checks all coin spends of a mempool item for deduplication eligibility and
        provides the caller with the necessary information that allows it to perform
        identical spend aggregation on that mempool item if possible

        Args:
            bundle_coin_spends: the mempool item's coin spends data
            max_cost: the maximum limit when running for cost

        Returns:
            list[CoinSpend]: list of unique coin spends in this mempool item
            uint64: the cost we're saving by deduplicating eligible coins
            list[Coin]: list of unique additions in this mempool item

        Raises:
            ValueError to skip the mempool item we're currently in, if it's
            attempting to spend an eligible coin with a different solution than the
            one we're already deduplicating on.
        """
        cost_saving = 0
        unique_coin_spends: list[CoinSpend] = []
        unique_additions: list[Coin] = []
        # Map of coin ID to deduplication information
        new_dedup_spends: dict[bytes32, DedupCoinSpend] = {}
        # See if this item has coin spends that are eligible for deduplication
        for coin_id, spend_data in bundle_coin_spends.items():
            if not spend_data.eligible_for_dedup:
                unique_coin_spends.append(spend_data.coin_spend)
                unique_additions.extend(spend_data.additions)
                continue
            # See if we processed an item with this coin before
            dedup_coin_spend = self.deduplication_spends.get(coin_id)
            if dedup_coin_spend is None:
                # We didn't process an item with this coin before. If we end up including
                # this item, add this pair to deduplication_spends
                new_dedup_spends[coin_id] = DedupCoinSpend(spend_data.coin_spend.solution, None)
                unique_coin_spends.append(spend_data.coin_spend)
                unique_additions.extend(spend_data.additions)
                continue
            # See if the solution was identical
            current_solution, duplicate_cost = dataclasses.astuple(dedup_coin_spend)
            if current_solution != spend_data.coin_spend.solution:
                # It wasn't, so let's skip this whole item because it's relying on
                # spending this coin with a different solution and that would
                # conflict with the coin spends that we're deduplicating already
                # NOTE: We can miss an opportunity to deduplicate on other solutions
                # even if they end up saving more cost, as we're going for the first
                # solution we see from the relatively highest FPC item, to avoid
                # severe performance and/or time-complexity impact
                raise ValueError("Solution is different from what we're deduplicating on")
            # Let's calculate the saved cost if we never did that before
            if duplicate_cost is None:
                # See first if this mempool item had this cost computed before
                # This can happen if this item didn't get included in the previous block
                spend_cost = spend_data.cost
                if spend_cost is None:
                    spend_cost = run_for_cost(
                        puzzle_reveal=spend_data.coin_spend.puzzle_reveal,
                        solution=spend_data.coin_spend.solution,
                        additions_count=len(spend_data.additions),
                        max_cost=max_cost,
                    )
                    # Update this mempool item's coin spends map
                    bundle_coin_spends[coin_id] = BundleCoinSpend(
                        coin_spend=spend_data.coin_spend,
                        eligible_for_dedup=spend_data.eligible_for_dedup,
                        eligible_for_fast_forward=spend_data.eligible_for_fast_forward,
                        additions=spend_data.additions,
                        cost=spend_cost,
                    )
                duplicate_cost = spend_cost
                # If we end up including this item, update this entry's cost
                new_dedup_spends[coin_id] = DedupCoinSpend(current_solution, duplicate_cost)
            cost_saving += duplicate_cost
        # Update the eligible coin spends data
        self.deduplication_spends.update(new_dedup_spends)
        return unique_coin_spends, uint64(cost_saving), unique_additions

    async def process_fast_forward_spends(
        self,
        *,
        mempool_item: InternalMempoolItem,
        get_unspent_lineage_info_for_puzzle_hash: Callable[[bytes32], Awaitable[Optional[UnspentLineageInfo]]],
        height: uint32,
        constants: ConsensusConstants,
    ) -> dict[bytes32, BundleCoinSpend]:
        """
        Provides the caller with a `bundle_coin_spends` map that has a proper
        state of fast forwarded coin spends and additions starting from
        the most recent unspent versions of the related singleton spends.

        Args:
            mempool_item: The internal mempool item to process
            get_unspent_lineage_info_for_puzzle_hash: to lookup the most recent
                version of the singleton from the coin store
            constants: needed in order to refresh the mempool item if needed
            height: needed in order to refresh the mempool item if needed

        Returns:
            The resulting `bundle_coin_spends` map of coin IDs to coin spends
            and metadata, after fast forwarding

        Raises:
            If a fast forward cannot proceed, to prevent potential double spends
        """

        # Let's first create a copy of the mempool item's `bundle_coin_spends`
        # map to work on and return. This way we avoid the possibility of
        # propagating a modified version of this item through the network.
        bundle_coin_spends = copy.copy(mempool_item.bundle_coin_spends)
        new_coin_spends = []
        # Map of rebased singleton coin ID to coin spend and metadata
        ff_bundle_coin_spends = {}
        replaced_coin_ids = []
        for coin_id, spend_data in bundle_coin_spends.items():
            if not spend_data.eligible_for_fast_forward:
                # Nothing to do for this spend, moving on
                new_coin_spends.append(spend_data.coin_spend)
                continue

            # NOTE: We need to support finding the most recent version of a singleton
            # both in the DB and in the current state of the block we are
            # building, in case we have already spent the singleton

            # See if we added a fast forward spend with this puzzle hash before
            unspent_lineage_info = self.fast_forward_spends.get(spend_data.coin_spend.coin.puzzle_hash)
            if unspent_lineage_info is None:
                # We didn't, so let's lookup the most recent version from the DB
                unspent_lineage_info = await get_unspent_lineage_info_for_puzzle_hash(
                    spend_data.coin_spend.coin.puzzle_hash
                )
                if unspent_lineage_info is None:
                    raise ValueError("Cannot proceed with singleton spend fast forward.")
                # See if we're the most recent version
                if unspent_lineage_info.coin_id == coin_id:
                    # We are, so we don't need to fast forward, we just need to
                    # set the next version from our additions to chain ff spends
                    set_next_singleton_version(
                        current_singleton=spend_data.coin_spend.coin,
                        singleton_additions=spend_data.additions,
                        fast_forward_spends=self.fast_forward_spends,
                    )
                    # Nothing more to do for this spend, moving on
                    new_coin_spends.append(spend_data.coin_spend)
                    continue
                # We're not the most recent version, so let's fast forward
                new_coin_spend, patched_additions = perform_the_fast_forward(
                    unspent_lineage_info=unspent_lineage_info,
                    spend_data=spend_data,
                    fast_forward_spends=self.fast_forward_spends,
                )
                # Mark this coin for a coin spend data update
                replaced_coin_ids.append(coin_id)
                ff_bundle_coin_spends[new_coin_spend.coin.name()] = BundleCoinSpend(
                    coin_spend=new_coin_spend,
                    eligible_for_dedup=spend_data.eligible_for_dedup,
                    eligible_for_fast_forward=spend_data.eligible_for_fast_forward,
                    additions=patched_additions,
                    cost=spend_data.cost,
                )
                # Update the list of coins spends that will make the new fast
                # forward spend bundle
                new_coin_spends.append(new_coin_spend)
                # We're done here, moving on
                continue
            # We've added a ff spend with this puzzle hash before, so build on that
            # NOTE: As it's not possible to submit a transaction to the mempool that
            # spends the output of another transaction already in the mempool,
            # we don't need to check if we're the most recent version because
            # at this point we cannot be, so we must fast forward
            new_coin_spend, patched_additions = perform_the_fast_forward(
                unspent_lineage_info=unspent_lineage_info,
                spend_data=spend_data,
                fast_forward_spends=self.fast_forward_spends,
            )
            # Mark this coin for a coin spend data update
            replaced_coin_ids.append(coin_id)
            ff_bundle_coin_spends[new_coin_spend.coin.name()] = BundleCoinSpend(
                coin_spend=new_coin_spend,
                eligible_for_dedup=spend_data.eligible_for_dedup,
                eligible_for_fast_forward=spend_data.eligible_for_fast_forward,
                additions=patched_additions,
                cost=spend_data.cost,
            )
            # Update the list of coins spends that make the new fast forward bundle
            new_coin_spends.append(new_coin_spend)
        if len(ff_bundle_coin_spends) == 0:
            # This item doesn't have any fast forward coins, nothing to do here
            return bundle_coin_spends
        # Update the mempool item after validating the new spend bundle
        new_sb = SpendBundle(
            coin_spends=new_coin_spends, aggregated_signature=mempool_item.spend_bundle.aggregated_signature
        )
        assert mempool_item.conds is not None
        try:
            # Run the new spend bundle to make sure it remains valid. What we
            # care about here is whether this call throws or not.
            get_conditions_from_spendbundle(new_sb, mempool_item.conds.cost, constants, height)
        # get_conditions_from_spendbundle raises a TypeError with an error code
        except TypeError as e:
            # Convert that to a ValidationError
            if len(e.args) > 0:
                error = Err(e.args[0])
                raise ValueError(f"Mempool item became invalid after singleton fast forward with error {error}.")
            else:
                raise ValueError(
                    "Mempool item became invalid after singleton fast forward with an unspecified error."
                )  # pragma: no cover

        # Update bundle_coin_spends using the map of rebased singleton coin ID
        # to coin spend and metadata.
        for coin_id in replaced_coin_ids:
            bundle_coin_spends.pop(coin_id, None)
        bundle_coin_spends.update(ff_bundle_coin_spends)
        return bundle_coin_spends



from __future__ import annotations

import pytest

from chik._tests.environments.wallet import (
    BalanceCheckingError,
    WalletEnvironment,
    WalletStateTransition,
    WalletTestFramework,
)
from chik.consensus.block_record import BlockRecord
from chik.wallet.cat_wallet.cat_wallet import CATWallet


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 2,
            "config_overrides": {
                "foo": "bar",  # A config value that never exists
                "min_mainnet_k_size": 2,  # A config value overriden
            },
            "blocks_needed": [1, 0],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_basic_functionality(wallet_environments: WalletTestFramework) -> None:
    env_0: WalletEnvironment = wallet_environments.environments[0]
    env_1: WalletEnvironment = wallet_environments.environments[1]

    assert (await env_0.rpc_client.get_logged_in_fingerprint()).fingerprint is not None
    # assert await env_1.rpc_client.get_logged_in_fingerprint() is not None

    assert await env_0.xck_wallet.get_confirmed_balance() == 2_000_000_000_000
    assert await env_1.xck_wallet.get_confirmed_balance() == 0

    assert env_0.node.config["foo"] == "bar"
    assert env_0.wallet_state_manager.config["foo"] == "bar"
    assert wallet_environments.full_node.full_node.config["foo"] == "bar"

    assert env_0.node.config["min_mainnet_k_size"] == 2
    assert env_0.wallet_state_manager.config["min_mainnet_k_size"] == 2
    assert wallet_environments.full_node.full_node.config["min_mainnet_k_size"] == 2

    assert isinstance(await wallet_environments.full_node_rpc_client.get_block_record_by_height(1), BlockRecord)


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "config_overrides": {},
            "blocks_needed": [1],
            "trusted": True,
            "reuse_puzhash": True,
        },
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_balance_checking(
    wallet_environments: WalletTestFramework,
) -> None:
    env_0: WalletEnvironment = wallet_environments.environments[0]
    await env_0.check_balances()
    await wallet_environments.full_node.farm_blocks_to_wallet(count=1, wallet=env_0.xck_wallet)
    await wallet_environments.full_node.wait_for_wallet_synced(wallet_node=env_0.node, timeout=20)
    with pytest.raises(BalanceCheckingError, match="2000000000000 compared to balance response 4000000000000"):
        await env_0.check_balances()
    with pytest.raises(KeyError):
        await env_0.change_balances(
            {
                "xck": {
                    "confirmed_wallet_balance": 2_000_000_000_000,
                    "unconfirmed_wallet_balance": 2_000_000_000_000,
                    "spendable_balance": 2_000_000_000_000,
                    "max_send_amount": 2_000_000_000_000,
                    "unspent_coin_count": 2,
                }
            }
        )
    env_0.wallet_aliases = {
        "xck": 1,
        "cat": 2,
    }
    with pytest.raises(ValueError, match="Error before block was farmed"):
        await wallet_environments.process_pending_states([WalletStateTransition()])

    with pytest.raises(ValueError, match="Error after block was farmed"):
        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        "xck": {
                            "confirmed_wallet_balance": 2_000_000_000_000,
                            "unconfirmed_wallet_balance": 2_000_000_000_000,
                            "spendable_balance": 2_000_000_000_000,
                            "max_send_amount": 2_000_000_000_000,
                            "unspent_coin_count": 2,
                        }
                    },
                    post_block_balance_updates={
                        "xck": {
                            "confirmed_wallet_balance": 13,
                        }
                    },
                )
            ]
        )

    # this is necessary to undo the changes made before raising above
    await env_0.change_balances(
        {
            "xck": {
                "confirmed_wallet_balance": -2_000_000_000_013,
                "unconfirmed_wallet_balance": -2_000_000_000_000,
                "spendable_balance": -2_000_000_000_000,
                "max_send_amount": -2_000_000_000_000,
                "unspent_coin_count": -2,
            }
        }
    )

    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    "xck": {
                        "confirmed_wallet_balance": 2_000_000_000_000,
                        "unconfirmed_wallet_balance": 2_000_000_000_000,
                        "spendable_balance": 2_000_000_000_000,
                        "max_send_amount": 2_000_000_000_000,
                        "unspent_coin_count": 2,
                    }
                },
                post_block_balance_updates={
                    "xck": {
                        "set_remainder": True,
                    }
                },
            )
        ]
    )
    await CATWallet.get_or_create_wallet_for_cat(env_0.wallet_state_manager, env_0.xck_wallet, "00" * 32)
    with pytest.raises(KeyError, match="No wallet state for wallet id 2"):
        await env_0.check_balances()

    with pytest.raises(ValueError, match="if you intended to initialize its state"):
        await env_0.change_balances(
            {
                "cat": {
                    "confirmed_wallet_balance": 1_000_000,
                    "unconfirmed_wallet_balance": 0,
                    "spendable_balance": 0,
                    "pending_change": 0,
                    "max_send_amount": 0,
                    "unspent_coin_count": 0,
                    "pending_coin_removal_count": 0,
                }
            }
        )

    await env_0.change_balances(
        {
            "cat": {
                "init": True,
                "confirmed_wallet_balance": 1_000_000,
                "unconfirmed_wallet_balance": 0,
                "spendable_balance": 0,
                "pending_change": 0,
                "max_send_amount": 0,
                "unspent_coin_count": 0,
                "pending_coin_removal_count": 0,
            }
        }
    )

    with pytest.raises(BalanceCheckingError):
        await env_0.check_balances()

    # Test override
    await env_0.check_balances(additional_balance_info={"cat": {"confirmed_wallet_balance": 0}})
    with pytest.raises(BalanceCheckingError, match="not in balance response"):
        await env_0.check_balances(additional_balance_info={"xck": {"something not there": 0}})

    await env_0.change_balances({"cat": {"init": True, "set_remainder": True}})
    await env_0.check_balances()

    # Test special operators
    await wallet_environments.full_node.farm_blocks_to_wallet(count=1, wallet=env_0.xck_wallet)
    await wallet_environments.full_node.wait_for_wallet_synced(wallet_node=env_0.node, timeout=20)
    with pytest.raises(ValueError, match=r"\+ 2000000000000"):
        await env_0.change_balances(
            {
                "xck": {
                    "<#confirmed_wallet_balance": 2_000_000_000_000,
                }
            }
        )

    with pytest.raises(ValueError, match=r"\+ 2000000000000"):
        await env_0.change_balances(
            {
                "xck": {
                    ">#confirmed_wallet_balance": 2_000_000_000_000,
                }
            }
        )

    with pytest.raises(ValueError, match=r"\+ 1999999999999"):
        await env_0.change_balances(
            {
                "xck": {
                    "<=#confirmed_wallet_balance": 1_999_999_999_999,
                }
            }
        )

    with pytest.raises(ValueError, match=r"\+ 2000000000001"):
        await env_0.change_balances(
            {
                "xck": {
                    ">=#confirmed_wallet_balance": 2_000_000_000_001,
                }
            }
        )

    wallet_states_save = env_0.wallet_states.copy()

    await env_0.change_balances(
        {
            "xck": {
                "<#confirmed_wallet_balance": 2_000_000_000_001,
                "set_remainder": True,
            }
        }
    )
    await env_0.check_balances()
    env_0.wallet_states = wallet_states_save

    await env_0.change_balances(
        {
            "xck": {
                ">#confirmed_wallet_balance": 1_999_999_999_999,
                "set_remainder": True,
            }
        }
    )
    await env_0.check_balances()
    env_0.wallet_states = wallet_states_save

    await env_0.change_balances(
        {
            "xck": {
                ">=#confirmed_wallet_balance": 2_000_000_000_000,
                "set_remainder": True,
            }
        }
    )
    await env_0.check_balances()
    env_0.wallet_states = wallet_states_save

    await env_0.change_balances(
        {
            "xck": {
                "<=#confirmed_wallet_balance": 2_000_000_000_000,
                "set_remainder": True,
            }
        }
    )
    await env_0.check_balances()



from __future__ import annotations

import json
import os
import re
from pathlib import Path
from typing import Optional

import pytest
from click.testing import CliRunner, Result

from chik.cmds.chik import cli
from chik.cmds.keys import delete_all_cmd, generate_and_print_cmd, sign_cmd, verify_cmd
from chik.cmds.keys_funcs import get_private_key_with_fingerprint_or_prompt
from chik.util.config import load_config
from chik.util.default_root import DEFAULT_KEYS_ROOT_PATH
from chik.util.keychain import Keychain, KeyData, generate_mnemonic
from chik.util.keyring_wrapper import KeyringWrapper

TEST_MNEMONIC_SEED = (
    "grief lock ketchup video day owner torch young work "
    "another venue evidence spread season bright private "
    "tomato remind jaguar original blur embody project can"
)
TEST_PUBLIC_KEY = "8a37cad4c5edf0a7544cbdb9f9383f7c5e82567d0236dc4f5b0547137afff9a5ce33aece3358d0202cafb9a12607ab53"
TEST_BECH32_PUBKEY = "bls123814rygqxr4ryyadg0tx9cprt55vzpamcumr34v3ys66cynl6qx43emhm35nlz8clfxnujzkhpvkvqf5n2fxpd"
TEST_PK_FINGERPRINT = 2167729070
TEST_FINGERPRINT = 2877570395


@pytest.fixture(scope="function")
def keyring_with_one_public_one_private_key(empty_keyring):
    keychain = empty_keyring
    keychain.add_key(TEST_MNEMONIC_SEED)
    keychain.add_key(TEST_PUBLIC_KEY, private=False)
    return keychain


@pytest.fixture(scope="function")
def mnemonic_seed_file(tmp_path):
    seed_file = Path(tmp_path) / "seed.txt"
    with open(seed_file, "w") as f:
        f.write(TEST_MNEMONIC_SEED)
    return seed_file


@pytest.fixture(scope="function")
def setup_keyringwrapper(tmp_path):
    KeyringWrapper.cleanup_shared_instance()
    KeyringWrapper.set_keys_root_path(tmp_path)
    _ = KeyringWrapper.get_shared_instance()
    yield
    KeyringWrapper.cleanup_shared_instance()
    KeyringWrapper.set_keys_root_path(DEFAULT_KEYS_ROOT_PATH)


def assert_label(keychain: Keychain, label: Optional[str], index: int) -> None:
    all_keys = keychain.get_keys()
    assert len(all_keys) > index
    assert all_keys[index].label == label


class TestKeysCommands:
    def test_generate_with_new_config(self, tmp_path, empty_keyring):
        """
        Generate a new config and a new key. Verify that the config has
        the correct xck_target_address entries.
        """

        keychain = empty_keyring
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        # Generate the new config
        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 0

        # Generate a new key
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "generate",
            ],
            input="\n",
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        # Verify that the config has the correct xck_target_address entries
        address_matches = re.findall(r"xck1[^\n]+", result.output)
        assert len(address_matches) > 1
        address = address_matches[0]

        config: dict = load_config(tmp_path, "config.yaml")
        assert config["farmer"]["xck_target_address"] == address
        assert config["pool"]["xck_target_address"] == address

    def test_generate_with_existing_config(self, tmp_path, empty_keyring):
        """
        Generate a new key using an existing config. Verify that the config has
        the original xck_target_address entries.
        """

        keychain = empty_keyring
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        # Generate the new config
        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 0

        # Generate the first key
        runner = CliRunner()
        generate_result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "generate",
            ],
            input="\n",
            catch_exceptions=False,
        )

        assert generate_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        # Verify that the config has the correct xck_target_address entries
        address_matches = re.findall(r"xck1[^\n]+", generate_result.output)
        assert len(address_matches) > 1
        address = address_matches[0]

        existing_config: dict = load_config(tmp_path, "config.yaml")
        assert existing_config["farmer"]["xck_target_address"] == address
        assert existing_config["pool"]["xck_target_address"] == address

        # Generate the second key
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "generate",
            ],
            input="\n",
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 2

        # Verify that the config's xck_target_address entries have not changed
        config: dict = load_config(tmp_path, "config.yaml")
        assert config["farmer"]["xck_target_address"] == existing_config["farmer"]["xck_target_address"]
        assert config["pool"]["xck_target_address"] == existing_config["pool"]["xck_target_address"]

    @pytest.mark.parametrize(
        "cmd_params, label, input_str",
        [
            (["generate"], None, "\n"),
            (["generate", "-l", "key_0"], "key_0", None),
            (["generate", "--label", "key_0"], "key_0", None),
            (["generate", "-l", ""], None, None),
            (["generate", "--label", ""], None, None),
            (["generate"], "key_0", "key_0\n"),
            (["add"], None, f"{TEST_MNEMONIC_SEED}\n\n"),
            (["add"], "key_0", f"{TEST_MNEMONIC_SEED}\nkey_0\n"),
            (["add", "-l", "key_0"], "key_0", f"{TEST_MNEMONIC_SEED}\n"),
            (["add", "--label", "key_0"], "key_0", f"{TEST_MNEMONIC_SEED}\n"),
            (["add", "-l", ""], None, f"{TEST_MNEMONIC_SEED}\n"),
            (["add", "--label", ""], None, f"{TEST_MNEMONIC_SEED}\n"),
            (["add", "-l", "key_0"], "key_0", f"{TEST_PUBLIC_KEY}\n"),
        ],
    )
    def test_generate_and_add_label_parameter(
        self, cmd_params: list[str], label: Optional[str], input_str: Optional[str], tmp_path, empty_keyring
    ):
        keychain = empty_keyring
        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        runner = CliRunner()
        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # Run the command
        result = runner.invoke(
            cli,
            [*base_params, "keys", *cmd_params],
            catch_exceptions=False,
            input=input_str,
        )
        assert result.exit_code == 0
        # And make sure the label was set to the expected label
        assert_label(keychain, label, 0)

    def test_set_label(self, keyring_with_one_public_one_private_key, tmp_path):
        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        cmd_params = ["keys", "label", "set", "-f", TEST_FINGERPRINT]
        runner = CliRunner()

        def set_and_validate(label: str):
            result = runner.invoke(cli, [*base_params, *cmd_params, "-l", label], catch_exceptions=False)
            assert result.exit_code == 0
            assert result.output == f"label {label!r} assigned to {TEST_FINGERPRINT!r}\n"
            assert_label(keychain, label, 0)

        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # There should be no label for this key
        assert_label(keychain, None, 0)
        # Set a label
        set_and_validate("key_0")
        # Change the label
        set_and_validate("changed")

    def test_delete_label(self, keyring_with_one_public_one_private_key, tmp_path):
        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        cmd_params = ["keys", "label", "delete", "-f", TEST_FINGERPRINT]
        runner = CliRunner()
        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # There should be no label for this key
        assert_label(keychain, None, 0)
        # Set a label
        keychain.set_label(TEST_FINGERPRINT, "key_0")
        assert_label(keychain, "key_0", 0)
        # Delete the label
        result = runner.invoke(cli, [*base_params, *cmd_params], catch_exceptions=False)
        assert result.output == f"label removed for {TEST_FINGERPRINT!r}\n"
        assert_label(keychain, None, 0)

    def test_show_labels(self, empty_keyring, tmp_path):
        keychain = empty_keyring
        runner = CliRunner()
        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        cmd_params = ["keys", "label", "show"]
        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # Make sure the command works with no keys
        result = runner.invoke(cli, [*base_params, *cmd_params], catch_exceptions=False)
        assert result.output == "No keys are present in the keychain. Generate them with 'chik keys generate'\n"
        # Add 10 keys to the keychain, give every other a label
        keys = [KeyData.generate(f"key_{i}" if i % 2 == 0 else None) for i in range(10)]
        for key in keys:
            keychain.add_key(key.mnemonic_str(), key.label)
        # Make sure all 10 keys are printed correct
        result = runner.invoke(cli, [*base_params, *cmd_params], catch_exceptions=False)
        assert result.exit_code == 0
        lines = result.output.splitlines()[2:]  # Split into lines but drop the header
        fingerprints = [int(line.split("|")[1].strip()) for line in lines]
        labels = [line.split("|")[2].strip() for line in lines]
        assert len(fingerprints) == len(labels) == len(keys)
        for fingerprint, label, key in zip(fingerprints, labels, keys):
            assert fingerprint == key.fingerprint
            if key.label is None:
                assert label == "No label assigned"
            else:
                assert label == key.label

    def test_show(self, keyring_with_one_public_one_private_key, tmp_path):
        """
        Test that the `chik keys show` command shows the correct key.
        """

        keychain = keyring_with_one_public_one_private_key

        assert len(keychain.get_all_private_keys()) == 1
        assert len(keychain.get_all_public_keys()) == 2

        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        runner = CliRunner()
        cmd_params = ["keys", "show"]
        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # Run the command
        result = runner.invoke(cli, [*base_params, *cmd_params], catch_exceptions=False)

        # assert result.exit_code == 0
        assert result.output.find(f"Fingerprint: {TEST_FINGERPRINT}") != -1
        assert result.output.find(f"Fingerprint: {TEST_PK_FINGERPRINT}") != -1

        # Try with non_observer_derivation
        result = runner.invoke(cli, [*base_params, *cmd_params, "--non-observer-derivation"])
        assert result.output.find(f"Fingerprint: {TEST_FINGERPRINT}") != -1
        assert result.output.find(f"Fingerprint: {TEST_PK_FINGERPRINT}") != -1
        assert result.output.find("First wallet address (non-observer): N/A") != -1

        # Try with bech32 formatting
        result = runner.invoke(cli, [*base_params, *cmd_params, "--bech32m-prefix", "bls1238"])
        assert result.output.find(TEST_BECH32_PUBKEY) != -1

    def test_show_fingerprint(self, keyring_with_one_public_one_private_key, tmp_path):
        """
        Test that the `chik keys show --fingerprint` command shows the correct key.
        """

        keychain = keyring_with_one_public_one_private_key

        # add a key
        keychain.add_key(generate_mnemonic())
        assert len(keychain.get_all_private_keys()) == 2
        assert len(keychain.get_all_public_keys()) == 3

        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        runner = CliRunner()
        cmd_params = ["keys", "show", "--fingerprint", TEST_FINGERPRINT]
        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # Run the command
        result = runner.invoke(cli, [*base_params, *cmd_params], catch_exceptions=False)

        assert result.exit_code == 0
        fingerprints = [line for line in result.output.splitlines() if "Fingerprint:" in line]
        assert len(fingerprints) == 1
        assert str(TEST_FINGERPRINT) in fingerprints[0]

    def test_show_json(self, keyring_with_one_public_one_private_key, tmp_path):
        """
        Test that the `chik keys show --json` command shows the correct key.
        """

        keychain = keyring_with_one_public_one_private_key

        assert len(keychain.get_all_private_keys()) == 1
        assert len(keychain.get_all_public_keys()) == 2

        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        runner = CliRunner()
        cmd_params = ["keys", "show", "--json"]
        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # Run the command
        result = runner.invoke(cli, [*base_params, *cmd_params], catch_exceptions=False)

        json_result = json.loads(result.output)

        # assert result.exit_code == 0
        assert json_result["keys"][0]["fingerprint"] == TEST_FINGERPRINT
        assert json_result["keys"][1]["fingerprint"] == TEST_PK_FINGERPRINT

    def test_show_mnemonic(self, keyring_with_one_public_one_private_key, tmp_path):
        """
        Test that the `chik keys show --show-mnemonic-seed` command shows the key's mnemonic seed.
        """

        keychain = keyring_with_one_public_one_private_key

        assert len(keychain.get_all_private_keys()) == 1

        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        runner = CliRunner()
        cmd_params = ["keys", "show", "--show-mnemonic-seed"]
        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # Run the command
        result = runner.invoke(cli, [*base_params, *cmd_params], catch_exceptions=False)

        # assert result.exit_code == 0
        assert result.output.find(f"Fingerprint: {TEST_FINGERPRINT}") != -1
        assert result.output.find("Mnemonic seed (24 secret words):") != -1
        assert result.output.find(TEST_MNEMONIC_SEED) != -1
        assert result.output.find(f"Fingerprint: {TEST_PK_FINGERPRINT}") != -1
        assert result.output.find("Mnemonic seed (24 secret words):\nN/A") != -1

    def test_show_mnemonic_json(self, keyring_with_one_public_one_private_key, tmp_path):
        """
        Test that the `chik keys show --show-mnemonic-seed --json` command shows the key's mnemonic seed.
        """

        keychain = keyring_with_one_public_one_private_key

        assert len(keychain.get_all_private_keys()) == 1

        keys_root_path = keychain.keyring_wrapper.keys_root_path
        base_params = [
            "--root-path",
            os.fspath(tmp_path),
            "--keys-root-path",
            os.fspath(keys_root_path),
        ]
        runner = CliRunner()
        cmd_params = ["keys", "show", "--show-mnemonic-seed", "--json"]
        # Generate a new config
        assert runner.invoke(cli, [*base_params, "init"], catch_exceptions=False).exit_code == 0
        # Run the command
        result = runner.invoke(cli, [*base_params, *cmd_params], catch_exceptions=False)
        json_result = json.loads(result.output)

        # assert result.exit_code == 0
        assert json_result["keys"][0]["fingerprint"] == TEST_FINGERPRINT
        assert json_result["keys"][0]["mnemonic"] == TEST_MNEMONIC_SEED
        assert json_result["keys"][1]["fingerprint"] == TEST_PK_FINGERPRINT
        assert json_result["keys"][1]["mnemonic"] is None

    def test_add_interactive(self, tmp_path, empty_keyring):
        """
        Test adding a key from mnemonic seed using the interactive prompt.
        """

        keychain = empty_keyring
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 0

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "add",
            ],
            catch_exceptions=False,
            input=f"{TEST_MNEMONIC_SEED}\n\n",
        )

        assert result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

    def test_add_from_mnemonic_seed(self, tmp_path, empty_keyring, mnemonic_seed_file):
        """
        Test adding a key from a mnemonic seed file using the `--filename` flag.
        """

        keychain = empty_keyring
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 0

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "add",
                "--filename",
                os.fspath(mnemonic_seed_file),
            ],
            catch_exceptions=False,
            input="\n",
        )

        assert result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

    def test_delete(self, tmp_path, empty_keyring, mnemonic_seed_file):
        """
        Test deleting a key using the `--fingerprint` option.
        """

        keychain = empty_keyring
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 0

        runner = CliRunner()
        add_result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "add",
                "--filename",
                os.fspath(mnemonic_seed_file),
            ],
            catch_exceptions=False,
            input="\n",
        )

        assert add_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "delete",
                "--fingerprint",
                TEST_FINGERPRINT,
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 0

    def test_delete_all(self, empty_keyring):
        """
        Test deleting all keys from the keyring
        """

        keychain = empty_keyring

        assert len(keychain.get_all_private_keys()) == 0

        for i in range(5):
            mnemonic: str = generate_mnemonic()
            keychain.add_key(mnemonic)

        assert len(keychain.get_all_private_keys()) == 5

        runner = CliRunner()
        result = runner.invoke(delete_all_cmd, [], catch_exceptions=False)

        assert result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 0

    def test_generate_and_print(self):
        """
        Test the `chik keys generate_and_print` command.
        """

        runner = CliRunner()
        result = runner.invoke(generate_and_print_cmd, [], catch_exceptions=False)

        assert result.exit_code == 0
        assert result.output.find("Mnemonic (24 secret words):") != -1

    def test_sign(self, keyring_with_one_public_one_private_key):
        """
        Test the `chik keys sign` command.
        """

        message: str = "hello world"
        hd_path: str = "m/12381/8444/0/1"
        runner = CliRunner()
        result = runner.invoke(
            sign_cmd,
            ["--message", message, "--fingerprint", str(TEST_FINGERPRINT), "--hd_path", hd_path],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Public Key: 92f15caed8a5495faa7ec25a8af3f223438ef73c974b0aa81e788057b1154870f149739b2c2d0e"
                "736234baf9386f7f83"
            )
            != -1
        )
        assert (
            result.output.find(
                "Signature: a82e7d1b87d8c25a6ccac603194011d73f71fc76c17c1ce4ee53484f81874f116b1cb9dd991bcf9"
                "aa41c10beaab54a830fc6f7e5e25a9144f73e38a6fb852a87e36d80f575a6f84359144e6e9499ba9208912de55"
                "a1f7514cd8cfa166ae48e64"
            )
            != -1
        )

    def test_sign_non_observer(self, keyring_with_one_public_one_private_key):
        """
        Test the `chik keys sign` command with a non-observer key.
        """

        message: str = "hello world"
        hd_path: str = "m/12381n/8444n/0n/1n"
        runner = CliRunner()
        result = runner.invoke(
            sign_cmd,
            ["--message", message, "--fingerprint", str(TEST_FINGERPRINT), "--hd_path", hd_path],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Public Key: b5e383b8192dacff662455bdb3bbfc433f678f0d7ff7f118149e0d2ad39aa6d59ac4cb3662acf8"
                "e8307e66069d3a13cc"
            )
        ) != -1
        assert (
            result.output.find(
                "Signature: b5b3bc1417f67498748018a7ad2c95acfc5ae2dcd0d9dd0f3abfc7e3f047f2e6cf6c3e775b6caff"
                "a3e0baaadc2fe705a100cd4c961d6ff3c575c5c33683eb7b1e2dbbcaf37318227ae40ef8ccf57879a7818fad8f"
                "dc573d55c908be2611b8077"
            )
        ) != -1

    def test_sign_mnemonic_seed_file(self, empty_keyring, mnemonic_seed_file):
        """
        Test signing a message using a key imported from a mnemonic seed file.
        """

        message: str = "hello world"
        hd_path: str = "m/12381/8444/0/1"
        runner = CliRunner()
        result = runner.invoke(
            sign_cmd,
            [
                "--message",
                message,
                "--hd_path",
                hd_path,
                "--mnemonic-seed-filename",
                mnemonic_seed_file,
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Public Key: "
                "92f15caed8a5495faa7ec25a8af3f223438ef73c974b0aa81e788057b1154870f149739b2c2d0e736234baf9386f7f83"
            )
            != -1
        )
        assert (
            result.output.find(
                "Signature: a82e7d1b87d8c25a6ccac603194011d73f71fc76c17c1ce4ee53484f81874f116b1cb9dd991bcf"
                "9aa41c10beaab54a830fc6f7e5e25a9144f73e38a6fb852a87e36d80f575a6f84359144e6e9499ba9208912de"
                "55a1f7514cd8cfa166ae48e64"
            )
            != -1
        )

    def test_verify(self):
        """
        Test the `chik keys verify` command.
        """

        message: str = "hello world"
        signature: str = (
            "a82e7d1b87d8c25a6ccac603194011d73f71fc76c17c1ce4ee53484f81874f116b1cb9dd991bcf9aa41c10beaab54a83"
            "0fc6f7e5e25a9144f73e38a6fb852a87e36d80f575a6f84359144e6e9499ba9208912de55a1f7514cd8cfa166ae48e64"
        )
        public_key: str = (
            "92f15caed8a5495faa7ec25a8af3f223438ef73c974b0aa81e788057b1154870f149739b2c2d0e736234baf9386f7f83"
        )
        runner = CliRunner()
        result = runner.invoke(
            verify_cmd,
            ["--message", message, "--public_key", public_key, "--signature", signature],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert result.output.find("True") == 0

    def test_derive_search(self, tmp_path, keyring_with_one_public_one_private_key):
        """
        Test the `chik keys derive search` command, searching a public and private key
        """

        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "search",
                "--limit",
                "10",
                "--search-type",
                "all",
                "a4601f992f24047097a30854ef656382911575694439108723698972941e402d737c13df76fdf43597f7b3c2fa9ed27a",
                "028e33fa3f8caa3102c028f3bff6b6680e528d9a0c543c479ef0b0339060ef36",
                "--show-progress",
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert result.output.find(f"Searching keys derived from: {TEST_FINGERPRINT}") != -1
        assert (
            result.output.find(
                "Found public key: a4601f992f24047097a30854ef656382911575694439108723698"
                "972941e402d737c13df76fdf43597f7b3c2fa9ed27a (HD path: m/12381/8444/2/9)"
            )
            != -1
        )
        assert (
            result.output.find(
                "Found private key: "
                "028e33fa3f8caa3102c028f3bff6b6680e528d9a0c543c479ef0b0339060ef36 (HD path: m/12381/8444/2/9)"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "search",
                "--limit",
                "10",
                "--search-type",
                "all",
                "83062a1b26d27820600eac4e31c1a890a6ba026b28bb96bb66454e9ce1033f4cba8824259dc17dc3b643ab1003e6b961",
                "--show-progress",
                "--non-observer-derivation",
            ],
        )

        assert result.exit_code == 0
        assert result.output.find(f"Searching keys derived from: {TEST_FINGERPRINT}") != -1
        assert (
            result.output.find(
                "Found public key: 83062a1b26d27820600eac4e31c1a890a6ba026b28bb96bb66454"
                "e9ce1033f4cba8824259dc17dc3b643ab1003e6b961 (HD path: m/12381n/8444n/2n/9n)"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "search",
                "--limit",
                "10",
                "--search-type",
                "private_key",
                "028e33fa3f8caa3102c028f3bff6b6680e528d9a0c543c479ef0b0339060ef36",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Found private key: 028e33fa3f8caa3102c028f3bff6b6680e528d9a0c543c479ef0b0339060ef36"
                " (HD path: m/12381/8444/2/9)"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "search",
                "--limit",
                "10",
                "--search-type",
                "all",
                "a272d5aaa6046e64bd7fd69bae288b9f9e5622c13058ec7d1b85e3d4d1acfa5d63d6542336c7b24d2fceab991919e989",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Found public key: a272d5aaa6046e64bd7fd69bae288b9f9e5622c13058ec7d1b85e"
                "3d4d1acfa5d63d6542336c7b24d2fceab991919e989 (HD path: m/12381/8444/2/9)"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "search",
                "--limit",
                "10",
                "--search-type",
                "all",
                "a272d5aaa6046e64bd7fd69bae288b9f9e5622c13058ec7d1b85e3d4d1acfa5d63d6542336c7b24d2fceab991919e989",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Found public key: a272d5aaa6046e64bd7fd69bae288b9f9e5622c13058ec7d1b85e"
                "3d4d1acfa5d63d6542336c7b24d2fceab991919e989 (HD path: m/12381/8444/2/9)"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "search",
                "--limit",
                "10",
                "--search-type",
                "all",
                "83062a1b26d27820600eac4e31c1a890a6ba026b28bb96bb66454e9ce1033f4cba8824259dc17dc3b643ab1003e6b961",
                "a272d5aaa6046e64bd7fd69bae288b9f9e5622c13058ec7d1b85e3d4d1acfa5d63d6542336c7b24d2fceab991919e989",
                "--non-observer-derivation",
            ],
        )

        assert result.exit_code == 1
        assert (
            result.output.find(
                "Found public key: 83062a1b26d27820600eac4e31c1a890a6ba026b28bb96bb66454"
                "e9ce1033f4cba8824259dc17dc3b643ab1003e6b961 (HD path: m/12381n/8444n/2n/9n)"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "search",
                "--limit",
                "10",
                "--search-type",
                "all",
                "a4601f992f24047097a30854ef656382911575694439108723698972941e402d737c13df76fdf43597f7b3c2fa9ed27a",
                "--show-progress",
            ],
        )

        assert result.exit_code == 1
        assert result.output.find(f"Searching keys derived from: {TEST_PK_FINGERPRINT}") != -1
        assert (
            result.output.find(
                "Could not find 'a4601f992f24047097a30854ef656382911575694439108723698972941e402d73"
                "7c13df76fdf43597f7b3c2fa9ed27a'"
            )
            != -1
        )

        # 83062a1b26d27820600eac4e31c1a890a6ba026b28bb96bb66454e9ce1033f4cba8824259dc17dc3b643ab1003e6b961

    def test_derive_search_wallet_address(self, tmp_path, keyring_with_one_public_one_private_key):
        """
        Test the `chik keys derive search` command, searching for a wallet address
        """

        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "search",
                "--limit",
                "40",
                "--search-type",
                "address",
                "xck1mnr0ygu7lvmk3nfgzmncfk39fwu0dv933yrcv97nd6pmrt7fzmhs0kyx8x",
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Found wallet address: "
                "xck1mnr0ygu7lvmk3nfgzmncfk39fwu0dv933yrcv97nd6pmrt7fzmhs0kyx8x (HD path: m/12381/8444/2/30)"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "search",
                "--limit",
                "40",
                "--search-type",
                "address",
                "xck1p33y7kv48u7l68m490mr8levl6nkyxm3x8tfcnnec555egxzd3gs0hupc5",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Found wallet address: "
                "xck1p33y7kv48u7l68m490mr8levl6nkyxm3x8tfcnnec555egxzd3gs0hupc5 (HD path: m/12381/8444/2/9)"
            )
            != -1
        )

    def test_derive_search_wallet_testnet_address(self, tmp_path, keyring_with_one_public_one_private_key):
        """
        Test the `chik keys derive search` command, searching for a testnet wallet address
        """

        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "search",
                "--limit",
                "40",
                "--search-type",
                "address",
                "txck1mnr0ygu7lvmk3nfgzmncfk39fwu0dv933yrcv97nd6pmrt7fzmhsz3rsx4",
                "--prefix",
                "txck",
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Found wallet address: "
                "txck1mnr0ygu7lvmk3nfgzmncfk39fwu0dv933yrcv97nd6pmrt7fzmhsz3rsx4 (HD path: m/12381/8444/2/30)"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "search",
                "--limit",
                "40",
                "--search-type",
                "address",
                "txck1p33y7kv48u7l68m490mr8levl6nkyxm3x8tfcnnec555egxzd3gszsmhe8",
                "--prefix",
                "txck",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Found wallet address: "
                "txck1p33y7kv48u7l68m490mr8levl6nkyxm3x8tfcnnec555egxzd3gszsmhe8 (HD path: m/12381/8444/2/9)"
            )
            != -1
        )

    def test_derive_search_failure(self, tmp_path, keyring_with_one_public_one_private_key):
        """
        Test the `chik keys derive search` command with a failing search.
        """

        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "search",
                "--limit",
                "10",
                "--search-type",
                "all",
                "something_that_doesnt_exist",
            ],
            catch_exceptions=False,
        )

        assert result.exit_code != 0

    def test_derive_search_hd_path(self, tmp_path, empty_keyring, mnemonic_seed_file):
        """
        Test the `chik keys derive search` command, searching under a provided HD path.
        """

        keychain = empty_keyring
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 0

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--mnemonic-seed-filename",
                os.fspath(mnemonic_seed_file),
                "search",
                "--limit",
                "50",
                "--search-type",
                "all",
                "--derive-from-hd-path",
                "m/12381n/8444n/2/",
                "80dc3a2ea450eb09e24debe22e1b5934911ba530792ef0be361badebb168780bd328ff8d4655e5dd573d5bef4a340344",
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Found public key: 80dc3a2ea450eb09e24debe22e1b5934911ba530792ef0be361bad"
                "ebb168780bd328ff8d4655e5dd573d5bef4a340344 (HD path: m/12381n/8444n/2/35)"
            )
            != -1
        )

    def test_derive_wallet_address(self, tmp_path, keyring_with_one_public_one_private_key, mnemonic_seed_file):
        """
        Test the `chik keys derive wallet-address` command, generating a couple of wallet addresses.
        """

        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--mnemonic-seed-filename",
                str(mnemonic_seed_file),
                "wallet-address",
                "--index",
                "50",
                "--count",
                "2",
                "--non-observer-derivation",
                "--show-hd-path",
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Wallet address 50 (m/12381n/8444n/2n/50n): "
                "xck1jp2u7an0mn9hdlw2x05nmje49gwgzmqyvh0qmh6008yksetuvkfsjn6xrt"
            )
            != -1
        )
        assert (
            result.output.find(
                "Wallet address 51 (m/12381n/8444n/2n/51n): "
                "xck1006n6l3x5e8exar8mlj004znjl5pq0tq73h76kz0yergswnjzn8s5x4x4q"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "wallet-address",
                "--index",
                "9",
                "--count",
                "1",
                "--show-hd-path",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Wallet address 9 (m/12381/8444/2/9): xck1p33y7kv48u7l68m490mr8levl6nkyxm3x8tfcnnec555egxzd3gs0hupc5"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "wallet-address",
                "--index",
                "9",
                "--count",
                "1",
                "--show-hd-path",
                "--non-observer-derivation",
            ],
        )
        assert result.exit_code == 0
        assert result.output.find("Could not resolve private key for non-observer derivation") != -1

    def test_derive_wallet_testnet_address(self, tmp_path, keyring_with_one_public_one_private_key):
        """
        Test the `chik keys derive wallet-address` command, generating a couple of testnet wallet addresses.
        """

        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "wallet-address",
                "--index",
                "50",
                "--count",
                "2",
                "--non-observer-derivation",
                "--show-hd-path",
                "--prefix",
                "txck",
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Wallet address 50 (m/12381n/8444n/2n/50n): "
                "txck1jp2u7an0mn9hdlw2x05nmje49gwgzmqyvh0qmh6008yksetuvkfsl5aszc"
            )
            != -1
        )
        assert (
            result.output.find(
                "Wallet address 51 (m/12381n/8444n/2n/51n): "
                "txck1006n6l3x5e8exar8mlj004znjl5pq0tq73h76kz0yergswnjzn8sepjs5n"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "wallet-address",
                "--index",
                "9",
                "--count",
                "1",
                "--show-hd-path",
                "--prefix",
                "txck",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Wallet address 9 (m/12381/8444/2/9): "
                "txck1p33y7kv48u7l68m490mr8levl6nkyxm3x8tfcnnec555egxzd3gszsmhe8"
            )
            != -1
        )

    def test_derive_child_keys(self, tmp_path, keyring_with_one_public_one_private_key, mnemonic_seed_file):
        """
        Test the `chik keys derive child-keys` command, generating a couple of derived keys.
        """

        keychain = keyring_with_one_public_one_private_key
        keys_root_path = keychain.keyring_wrapper.keys_root_path

        runner = CliRunner()
        init_result = runner.invoke(
            cli,
            ["--root-path", os.fspath(tmp_path), "--keys-root-path", os.fspath(keys_root_path), "init"],
            catch_exceptions=False,
        )

        assert init_result.exit_code == 0
        assert len(keychain.get_all_private_keys()) == 1

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--mnemonic-seed-filename",
                str(mnemonic_seed_file),
                "child-key",
                "--derive-from-hd-path",
                "m/12381n/8444n/2/3/4/",
                "--index",
                "30",
                "--count",
                "2",
                "--show-private-keys",
                "--show-hd-path",
            ],
            catch_exceptions=False,
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Observer public key 30 (m/12381n/8444n/2/3/4/30): "
                "979a1fa0bfc140488d4a9edcfbf244a398fe922618a981cc0fffe5445d811f2237ff8234c0520b28b3096c8269f2731e"
            )
            != -1
        )
        assert (
            result.output.find(
                "Observer private key 30 (m/12381n/8444n/2/3/4/30): "
                "5dd22db24fe28805b101104c543f5bec3808328ad67de3d3dcd9efd6faab13aa"
            )
            != -1
        )
        assert (
            result.output.find(
                "Observer public key 31 (m/12381n/8444n/2/3/4/31): "
                "ab5885df340a27b5eb3f1c4b8c32889f529ad5ecc4c9718247e36756de2e143c604af9956941a72239124e6fb352782e"
            )
            != -1
        )
        assert (
            result.output.find(
                "Observer private key 31 (m/12381n/8444n/2/3/4/31): "
                "113610b39c2151fd68d7f795d5dd596b94889a3cf7825a56da5c6d2c7e5141a1"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "child-key",
                "--type",
                "wallet",
                "--index",
                "9",
                "--count",
                "1",
                "--show-hd-path",
                "--show-private-keys",
                "--non-observer-derivation",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Wallet public key 9 (m/12381n/8444n/2n/9n): "
                "83062a1b26d27820600eac4e31c1a890a6ba026b28bb96bb66454e9ce1033f4cba8824259dc17dc3b643ab1003e6b961"
            )
            != -1
        )
        assert (
            result.output.find(
                "Wallet private key 9 (m/12381n/8444n/2n/9n): "
                "522f45786db6446d2f617a0c7df894385a21d05c7fbbfb34ee5aaaa417d8f41f"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_FINGERPRINT),
                "child-key",
                "--type",
                "wallet",
                "--index",
                "9",
                "--count",
                "1",
                "--show-hd-path",
                "--show-private-keys",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Wallet public key 9 (m/12381/8444/2/9): "
                "a4601f992f24047097a30854ef656382911575694439108723698972941e402d737c13df76fdf43597f7b3c2fa9ed27a"
            )
            != -1
        )
        assert (
            result.output.find(
                "Wallet private key 9 (m/12381/8444/2/9): "
                "028e33fa3f8caa3102c028f3bff6b6680e528d9a0c543c479ef0b0339060ef36"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "child-key",
                "--derive-from-hd-path",
                "m/12381/8444/2/3/4/",
                "--index",
                "30",
                "--count",
                "2",
                "--show-hd-path",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Observer public key 30 (m/12381/8444/2/3/4/30): "
                "b64b1c85bacc8fe3509a559d0178ce8373d8ea1dfcdb0c4b76fa03540da31d01276f7f000306d576c1c29468cf13b45a"
            )
            != -1
        )
        assert (
            result.output.find(
                "Observer public key 31 (m/12381/8444/2/3/4/31): "
                "92ba067175f2e87dd47336302ef8331a05ee0a737a92deec1b23dcd49eeb783d7feafdd055e3cba24fd6c94b39eb7bf3"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "child-key",
                "--type",
                "wallet",
                "--index",
                "9",
                "--count",
                "1",
                "--show-hd-path",
                "--bech32m-prefix",
                "bls1238",
            ],
        )

        assert result.exit_code == 0
        assert (
            result.output.find(
                "Wallet public key 9 (m/12381/8444/2/9): "
                "bls123815fedt24xq3hxf0tl66d6u2ytn709vgkpxpvwclgmsh3af5dvlfwk84j5yvmv0vjd9l82hxger85cjly2r27"
            )
            != -1
        )

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "child-key",
                "--type",
                "wallet",
                "--index",
                "9",
                "--count",
                "1",
                "--show-hd-path",
                "--non-observer-derivation",
            ],
        )

        assert result.output.find("Could not resolve private key for non-observer derivation") != -1

        result: Result = runner.invoke(
            cli,
            [
                "--root-path",
                os.fspath(tmp_path),
                "--keys-root-path",
                os.fspath(keys_root_path),
                "keys",
                "derive",
                "--fingerprint",
                str(TEST_PK_FINGERPRINT),
                "child-key",
                "--derive-from-hd-path",
                "m/12381n/8444/2/3/4/",
                "--index",
                "30",
                "--count",
                "2",
                "--show-hd-path",
            ],
        )

        assert isinstance(result.exception, ValueError) and result.exception.args == (
            "Hardened path specified for observer key",
        )

    @pytest.mark.anyio
    async def test_get_private_key_with_fingerprint_or_prompt(
        self, monkeypatch, keyring_with_one_public_one_private_key
    ) -> None:
        [sk1_plus_ent] = keyring_with_one_public_one_private_key.get_all_private_keys()
        sk1, _ = sk1_plus_ent
        [pk1, pk2] = keyring_with_one_public_one_private_key.get_all_public_keys()
        assert pk1.get_fingerprint() == TEST_FINGERPRINT
        assert pk2.get_fingerprint() == TEST_PK_FINGERPRINT

        assert get_private_key_with_fingerprint_or_prompt(TEST_FINGERPRINT) == (TEST_FINGERPRINT, sk1)
        assert get_private_key_with_fingerprint_or_prompt(TEST_PK_FINGERPRINT) == (TEST_PK_FINGERPRINT, None)

        monkeypatch.setattr("builtins.input", lambda _: "1")
        assert get_private_key_with_fingerprint_or_prompt(None) == (TEST_FINGERPRINT, sk1)

        monkeypatch.setattr("builtins.input", lambda _: "2")
        assert get_private_key_with_fingerprint_or_prompt(None) == (TEST_PK_FINGERPRINT, None)



from __future__ import annotations

import contextlib
import json
from collections.abc import AsyncIterator
from pathlib import Path
from typing import Any, Optional

from chik.cmds.cmds_util import get_any_service_client
from chik.rpc.data_layer_rpc_client import DataLayerRpcClient
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.byte_types import hexstr_to_bytes
from chik.util.default_root import resolve_root_path
from chik.util.ints import uint64


@contextlib.asynccontextmanager
async def get_client(
    rpc_port: Optional[int], fingerprint: Optional[int] = None, root_path: Optional[Path] = None
) -> AsyncIterator[tuple[DataLayerRpcClient, dict[str, Any]]]:
    root_path = resolve_root_path(override=root_path)

    async with get_any_service_client(
        client_type=DataLayerRpcClient,
        rpc_port=rpc_port,
        root_path=root_path,
    ) as (client, _):
        if fingerprint is not None:
            await client.wallet_log_in(fingerprint=fingerprint)
        yield client, _


async def wallet_log_in_cmd(
    rpc_port: Optional[int],
    fingerprint: int,
    root_path: Optional[Path] = None,
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path):
        pass


async def create_data_store_cmd(
    rpc_port: Optional[int],
    fee: Optional[uint64],
    verbose: bool,
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.create_data_store(fee=fee, verbose=verbose)
        print(json.dumps(res, indent=2, sort_keys=True))


async def get_value_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    key: str,
    # NOTE: being outside the rpc, this retains the none-means-unspecified semantics
    root_hash: Optional[bytes32],
    fingerprint: Optional[int],
) -> None:
    key_bytes = hexstr_to_bytes(key)
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.get_value(store_id=store_id, key=key_bytes, root_hash=root_hash)
        print(json.dumps(res, indent=2, sort_keys=True))


async def update_data_store_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    changelist: list[dict[str, str]],
    fee: Optional[uint64],
    fingerprint: Optional[int],
    submit_on_chain: bool,
    root_path: Optional[Path] = None,
) -> dict[str, Any]:
    res = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        res = await client.update_data_store(
            store_id=store_id,
            changelist=changelist,
            fee=fee,
            submit_on_chain=submit_on_chain,
        )
        print(json.dumps(res, indent=2, sort_keys=True))

    return res


async def update_multiple_stores_cmd(
    rpc_port: Optional[int],
    store_updates: list[dict[str, str]],
    fee: Optional[uint64],
    fingerprint: Optional[int],
    submit_on_chain: bool,
    root_path: Optional[Path] = None,
) -> dict[str, Any]:
    res = dict()

    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        res = await client.update_multiple_stores(
            store_updates=store_updates,
            fee=fee,
            submit_on_chain=submit_on_chain,
        )
        print(json.dumps(res, indent=2, sort_keys=True))

    return res


async def submit_pending_root_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    fee: Optional[uint64],
    fingerprint: Optional[int],
    root_path: Optional[Path] = None,
) -> dict[str, Any]:
    res = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        res = await client.submit_pending_root(
            store_id=store_id,
            fee=fee,
        )
        print(json.dumps(res, indent=2, sort_keys=True))

    return res


async def submit_all_pending_roots_cmd(
    rpc_port: Optional[int],
    fee: Optional[uint64],
    fingerprint: Optional[int],
    root_path: Optional[Path] = None,
) -> dict[str, Any]:
    res = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        res = await client.submit_all_pending_roots(fee=fee)
        print(json.dumps(res, indent=2, sort_keys=True))

    return res


async def get_keys_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    # NOTE: being outside the rpc, this retains the none-means-unspecified semantics
    root_hash: Optional[bytes32],
    fingerprint: Optional[int],
    page: Optional[int],
    max_page_size: Optional[int],
    root_path: Optional[Path] = None,
) -> dict[str, Any]:
    res = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        res = await client.get_keys(store_id=store_id, root_hash=root_hash, page=page, max_page_size=max_page_size)
        print(json.dumps(res, indent=2, sort_keys=True))

    return res


async def get_keys_values_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    # NOTE: being outside the rpc, this retains the none-means-unspecified semantics
    root_hash: Optional[bytes32],
    fingerprint: Optional[int],
    page: Optional[int],
    max_page_size: Optional[int],
    root_path: Optional[Path] = None,
) -> dict[str, Any]:
    res = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        res = await client.get_keys_values(
            store_id=store_id, root_hash=root_hash, page=page, max_page_size=max_page_size
        )
        print(json.dumps(res, indent=2, sort_keys=True))

    return res


async def get_root_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.get_root(store_id=store_id)
        print(json.dumps(res, indent=2, sort_keys=True))


async def subscribe_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    urls: list[str],
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.subscribe(store_id=store_id, urls=urls)
        print(json.dumps(res, indent=2, sort_keys=True))


async def unsubscribe_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    fingerprint: Optional[int],
    retain: bool,
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.unsubscribe(store_id=store_id, retain=retain)
        print(json.dumps(res, indent=2, sort_keys=True))


async def remove_subscriptions_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    urls: list[str],
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.remove_subscriptions(store_id=store_id, urls=urls)
        print(json.dumps(res, indent=2, sort_keys=True))


async def get_kv_diff_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    hash_1: bytes32,
    hash_2: bytes32,
    fingerprint: Optional[int],
    page: Optional[int],
    max_page_size: Optional[int],
    root_path: Optional[Path] = None,
) -> dict[str, Any]:
    res = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        res = await client.get_kv_diff(
            store_id=store_id, hash_1=hash_1, hash_2=hash_2, page=page, max_page_size=max_page_size
        )
        print(json.dumps(res, indent=2, sort_keys=True))

    return res


async def get_root_history_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.get_root_history(store_id=store_id)
        print(json.dumps(res, indent=2, sort_keys=True))


async def add_missing_files_cmd(
    rpc_port: Optional[int],
    ids: Optional[list[bytes32]],
    overwrite: bool,
    foldername: Optional[Path],
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.add_missing_files(
            store_ids=ids,
            overwrite=overwrite,
            foldername=foldername,
        )
        print(json.dumps(res, indent=2, sort_keys=True))


async def add_mirror_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    urls: list[str],
    amount: int,
    fee: Optional[uint64],
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.add_mirror(
            store_id=store_id,
            urls=urls,
            amount=amount,
            fee=fee,
        )
        print(json.dumps(res, indent=2, sort_keys=True))


async def delete_mirror_cmd(
    rpc_port: Optional[int],
    coin_id: bytes32,
    fee: Optional[uint64],
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.delete_mirror(
            coin_id=coin_id,
            fee=fee,
        )
        print(json.dumps(res, indent=2, sort_keys=True))


async def get_mirrors_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.get_mirrors(store_id=store_id)
        print(json.dumps(res, indent=2, sort_keys=True))


async def get_subscriptions_cmd(
    rpc_port: Optional[int],
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.get_subscriptions()
        print(json.dumps(res, indent=2, sort_keys=True))


async def get_owned_stores_cmd(
    rpc_port: Optional[int],
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.get_owned_stores()
        print(json.dumps(res, indent=2, sort_keys=True))


async def get_sync_status_cmd(
    rpc_port: Optional[int],
    store_id: bytes32,
    fingerprint: Optional[int],
) -> None:
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint) as (client, _):
        res = await client.get_sync_status(store_id=store_id)
        print(json.dumps(res, indent=2, sort_keys=True))


async def check_plugins_cmd(rpc_port: Optional[int]) -> None:
    async with get_client(rpc_port=rpc_port) as (client, _):
        res = await client.check_plugins()
        print(json.dumps(res, indent=2, sort_keys=True))


async def clear_pending_roots(
    store_id: bytes32,
    rpc_port: Optional[int],
    root_path: Optional[Path] = None,
    fingerprint: Optional[int] = None,
) -> dict[str, Any]:
    result = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        result = await client.clear_pending_roots(store_id=store_id)
        print(json.dumps(result, indent=2, sort_keys=True))

    return result


async def get_proof_cmd(
    store_id: bytes32,
    key_strings: list[str],
    rpc_port: Optional[int],
    root_path: Optional[Path] = None,
    fingerprint: Optional[int] = None,
) -> dict[str, Any]:
    result = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        result = await client.get_proof(store_id=store_id, keys=[hexstr_to_bytes(key) for key in key_strings])
        print(json.dumps(result, indent=2, sort_keys=True))

    return result


async def verify_proof_cmd(
    proof: dict[str, Any],
    rpc_port: Optional[int],
    root_path: Optional[Path] = None,
    fingerprint: Optional[int] = None,
) -> dict[str, Any]:
    result = dict()
    async with get_client(rpc_port=rpc_port, fingerprint=fingerprint, root_path=root_path) as (client, _):
        result = await client.verify_proof(proof=proof)
        print(json.dumps(result, indent=2, sort_keys=True))

    return result



from __future__ import annotations

from dataclasses import dataclass, field

from chik_rs import AugSchemeMPL, G1Element, G2Element, PrivateKey

from chik._tests.core.make_block_generator import GROUP_ORDER, int_to_public_key
from chik.simulator.block_tools import test_constants
from chik.types.coin_spend import CoinSpend
from chik.util.condition_tools import conditions_dict_for_solution, pkm_pairs_for_conditions_dict


@dataclass
class KeyTool:
    dict: dict[G1Element, int] = field(default_factory=dict)

    def add_secret_exponents(self, secret_exponents: list[int]) -> None:
        for _ in secret_exponents:
            self.dict[int_to_public_key(_)] = _ % GROUP_ORDER

    def sign(self, public_key: G1Element, message: bytes) -> G2Element:
        secret_exponent = self.dict.get(public_key)
        if not secret_exponent:
            raise ValueError(f"unknown pubkey {bytes(public_key).hex()}")
        bls_private_key = PrivateKey.from_bytes(secret_exponent.to_bytes(32, "big"))
        return AugSchemeMPL.sign(bls_private_key, message)

    def signature_for_solution(self, coin_spend: CoinSpend, additional_data: bytes) -> G2Element:
        signatures = []
        conditions_dict = conditions_dict_for_solution(
            coin_spend.puzzle_reveal, coin_spend.solution, test_constants.MAX_BLOCK_COST_KLVM
        )
        for public_key, message in pkm_pairs_for_conditions_dict(conditions_dict, coin_spend.coin, additional_data):
            signature = self.sign(public_key, message)
            signatures.append(signature)
        return AugSchemeMPL.aggregate(signatures)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional

if TYPE_CHECKING:
    from chik._tests.util.misc import TestId

# NOTE: Do not just put any useful thing here.  This is specifically for making
#       fixture values globally available during tests.  In _most_ cases fixtures
#       should be directly requested using normal mechanisms.  Very little should
#       be put here.

# NOTE: When using this module do not import the attributes directly.  Rather, import
#       something like `from chik._tests import ether`.  Importing attributes directly will
#       result in you likely getting the default `None` values since they are not
#       populated until tests are running.

record_property: Optional[Callable[[str, object], None]] = None
test_id: Optional[TestId] = None



from __future__ import annotations

from typing import Any, Optional, cast

from chik.consensus.block_record import BlockRecord
from chik.full_node.signage_point import SignagePoint
from chik.rpc.rpc_client import RpcClient
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_record import CoinRecord
from chik.types.coin_spend import CoinSpend, CoinSpendWithConditions
from chik.types.end_of_slot_bundle import EndOfSubSlotBundle
from chik.types.full_block import FullBlock
from chik.types.spend_bundle import SpendBundle
from chik.types.unfinished_header_block import UnfinishedHeaderBlock
from chik.util.ints import uint32


def coin_record_dict_backwards_compat(coin_record: dict[str, Any]) -> dict[str, Any]:
    del coin_record["spent"]
    return coin_record


class FullNodeRpcClient(RpcClient):
    """
    Client to Chik RPC, connects to a local full node. Uses HTTP/JSON, and converts back from
    JSON into native python objects before returning. All api calls use POST requests.
    Note that this is not the same as the peer protocol, or wallet protocol (which run Chik's
    protocol on top of TCP), it's a separate protocol on top of HTTP that provides easy access
    to the full node.
    """

    async def get_blockchain_state(self) -> dict[str, Any]:
        response = await self.fetch("get_blockchain_state", {})
        if response["blockchain_state"]["peak"] is not None:
            response["blockchain_state"]["peak"] = BlockRecord.from_json_dict(response["blockchain_state"]["peak"])
        return cast(dict[str, Any], response["blockchain_state"])

    async def get_block(self, header_hash: bytes32) -> Optional[FullBlock]:
        try:
            response = await self.fetch("get_block", {"header_hash": header_hash.hex()})
        except Exception:
            return None
        return FullBlock.from_json_dict(response["block"])

    async def get_blocks(self, start: int, end: int, exclude_reorged: bool = False) -> list[FullBlock]:
        response = await self.fetch(
            "get_blocks", {"start": start, "end": end, "exclude_header_hash": True, "exclude_reorged": exclude_reorged}
        )
        return [FullBlock.from_json_dict(block) for block in response["blocks"]]

    async def get_block_record_by_height(self, height: int) -> Optional[BlockRecord]:
        try:
            response = await self.fetch("get_block_record_by_height", {"height": height})
        except Exception:
            return None
        return BlockRecord.from_json_dict(response["block_record"])

    async def get_block_record(self, header_hash: bytes32) -> Optional[BlockRecord]:
        try:
            response = await self.fetch("get_block_record", {"header_hash": header_hash.hex()})
            if response["block_record"] is None:
                return None
        except Exception:
            return None
        return BlockRecord.from_json_dict(response["block_record"])

    async def get_unfinished_block_headers(self) -> list[UnfinishedHeaderBlock]:
        response = await self.fetch("get_unfinished_block_headers", {})
        return [UnfinishedHeaderBlock.from_json_dict(r) for r in response["headers"]]

    async def get_all_block(self, start: uint32, end: uint32) -> list[FullBlock]:
        response = await self.fetch("get_blocks", {"start": start, "end": end, "exclude_header_hash": True})
        return [FullBlock.from_json_dict(r) for r in response["blocks"]]

    async def get_network_space(self, newer_block_header_hash: bytes32, older_block_header_hash: bytes32) -> int:
        network_space_bytes_estimate = await self.fetch(
            "get_network_space",
            {
                "newer_block_header_hash": newer_block_header_hash.hex(),
                "older_block_header_hash": older_block_header_hash.hex(),
            },
        )

        return cast(int, network_space_bytes_estimate["space"])

    async def get_coin_record_by_name(self, coin_id: bytes32) -> Optional[CoinRecord]:
        try:
            response = await self.fetch("get_coin_record_by_name", {"name": coin_id.hex()})
        except Exception:
            return None

        return CoinRecord.from_json_dict(coin_record_dict_backwards_compat(response["coin_record"]))

    async def get_coin_records_by_names(
        self,
        names: list[bytes32],
        include_spent_coins: bool = True,
        start_height: Optional[int] = None,
        end_height: Optional[int] = None,
    ) -> list[CoinRecord]:
        names_hex = [name.hex() for name in names]
        d = {"names": names_hex, "include_spent_coins": include_spent_coins}
        if start_height is not None:
            d["start_height"] = start_height
        if end_height is not None:
            d["end_height"] = end_height

        response = await self.fetch("get_coin_records_by_names", d)
        return [CoinRecord.from_json_dict(coin_record_dict_backwards_compat(coin)) for coin in response["coin_records"]]

    async def get_coin_records_by_puzzle_hash(
        self,
        puzzle_hash: bytes32,
        include_spent_coins: bool = True,
        start_height: Optional[int] = None,
        end_height: Optional[int] = None,
    ) -> list[CoinRecord]:
        d = {"puzzle_hash": puzzle_hash.hex(), "include_spent_coins": include_spent_coins}
        if start_height is not None:
            d["start_height"] = start_height
        if end_height is not None:
            d["end_height"] = end_height

        response = await self.fetch("get_coin_records_by_puzzle_hash", d)
        return [CoinRecord.from_json_dict(coin_record_dict_backwards_compat(coin)) for coin in response["coin_records"]]

    async def get_coin_records_by_puzzle_hashes(
        self,
        puzzle_hashes: list[bytes32],
        include_spent_coins: bool = True,
        start_height: Optional[int] = None,
        end_height: Optional[int] = None,
    ) -> list[CoinRecord]:
        puzzle_hashes_hex = [ph.hex() for ph in puzzle_hashes]
        d = {"puzzle_hashes": puzzle_hashes_hex, "include_spent_coins": include_spent_coins}
        if start_height is not None:
            d["start_height"] = start_height
        if end_height is not None:
            d["end_height"] = end_height

        response = await self.fetch("get_coin_records_by_puzzle_hashes", d)
        return [CoinRecord.from_json_dict(coin_record_dict_backwards_compat(coin)) for coin in response["coin_records"]]

    async def get_coin_records_by_parent_ids(
        self,
        parent_ids: list[bytes32],
        include_spent_coins: bool = True,
        start_height: Optional[int] = None,
        end_height: Optional[int] = None,
    ) -> list[CoinRecord]:
        parent_ids_hex = [pid.hex() for pid in parent_ids]
        d = {"parent_ids": parent_ids_hex, "include_spent_coins": include_spent_coins}
        if start_height is not None:
            d["start_height"] = start_height
        if end_height is not None:
            d["end_height"] = end_height

        response = await self.fetch("get_coin_records_by_parent_ids", d)
        return [CoinRecord.from_json_dict(coin_record_dict_backwards_compat(coin)) for coin in response["coin_records"]]

    async def get_aggsig_additional_data(self) -> bytes32:
        result = await self.fetch("get_aggsig_additional_data", {})
        return bytes32.from_hexstr(result["additional_data"])

    async def get_coin_records_by_hint(
        self,
        hint: bytes32,
        include_spent_coins: bool = True,
        start_height: Optional[int] = None,
        end_height: Optional[int] = None,
    ) -> list[CoinRecord]:
        d = {"hint": hint.hex(), "include_spent_coins": include_spent_coins}
        if start_height is not None:
            d["start_height"] = start_height
        if end_height is not None:
            d["end_height"] = end_height

        response = await self.fetch("get_coin_records_by_hint", d)
        return [CoinRecord.from_json_dict(coin_record_dict_backwards_compat(coin)) for coin in response["coin_records"]]

    async def get_additions_and_removals(self, header_hash: bytes32) -> tuple[list[CoinRecord], list[CoinRecord]]:
        try:
            response = await self.fetch("get_additions_and_removals", {"header_hash": header_hash.hex()})
        except Exception:
            return [], []
        removals = []
        additions = []
        for coin_record in response["removals"]:
            removals.append(CoinRecord.from_json_dict(coin_record_dict_backwards_compat(coin_record)))
        for coin_record in response["additions"]:
            additions.append(CoinRecord.from_json_dict(coin_record_dict_backwards_compat(coin_record)))
        return additions, removals

    async def get_block_records(self, start: int, end: int) -> list[dict[str, Any]]:
        try:
            response = await self.fetch("get_block_records", {"start": start, "end": end})
            if response["block_records"] is None:
                return []
        except Exception:
            return []
        # TODO: return block records
        return cast(list[dict[str, Any]], response["block_records"])

    async def get_block_spends(self, header_hash: bytes32) -> Optional[list[CoinSpend]]:
        try:
            response = await self.fetch("get_block_spends", {"header_hash": header_hash.hex()})
            block_spends = []
            for block_spend in response["block_spends"]:
                block_spends.append(CoinSpend.from_json_dict(block_spend))
            return block_spends
        except Exception:
            return None

    async def get_block_spends_with_conditions(self, header_hash: bytes32) -> Optional[list[CoinSpendWithConditions]]:
        try:
            response = await self.fetch("get_block_spends_with_conditions", {"header_hash": header_hash.hex()})
            block_spends: list[CoinSpendWithConditions] = []
            for block_spend in response["block_spends_with_conditions"]:
                block_spends.append(CoinSpendWithConditions.from_json_dict(block_spend))
            return block_spends

        except Exception:
            return None

    async def push_tx(self, spend_bundle: SpendBundle) -> dict[str, Any]:
        return await self.fetch("push_tx", {"spend_bundle": spend_bundle.to_json_dict()})

    async def get_puzzle_and_solution(self, coin_id: bytes32, height: uint32) -> Optional[CoinSpend]:
        try:
            response = await self.fetch("get_puzzle_and_solution", {"coin_id": coin_id.hex(), "height": height})
            return CoinSpend.from_json_dict(response["coin_solution"])
        except Exception:
            return None

    async def get_all_mempool_tx_ids(self) -> list[bytes32]:
        response = await self.fetch("get_all_mempool_tx_ids", {})
        return [bytes32.from_hexstr(tx_id_hex) for tx_id_hex in response["tx_ids"]]

    async def get_all_mempool_items(self) -> dict[bytes32, dict[str, Any]]:
        response = await self.fetch("get_all_mempool_items", {})
        converted: dict[bytes32, dict[str, Any]] = {}
        for tx_id_hex, item in response["mempool_items"].items():
            converted[bytes32.from_hexstr(tx_id_hex)] = item
        return converted

    async def get_mempool_item_by_tx_id(
        self,
        tx_id: bytes32,
        include_pending: bool = False,
    ) -> Optional[dict[str, Any]]:
        try:
            response = await self.fetch(
                "get_mempool_item_by_tx_id", {"tx_id": tx_id.hex(), "include_pending": include_pending}
            )
            return cast(dict[str, Any], response["mempool_item"])
        except Exception:
            return None

    async def get_mempool_items_by_coin_name(self, coin_name: bytes32) -> dict[str, Any]:
        response = await self.fetch("get_mempool_items_by_coin_name", {"coin_name": coin_name.hex()})
        return response

    async def get_recent_signage_point_or_eos(
        self, sp_hash: Optional[bytes32], challenge_hash: Optional[bytes32]
    ) -> Optional[Any]:
        try:
            if sp_hash is not None:
                assert challenge_hash is None
                response = await self.fetch("get_recent_signage_point_or_eos", {"sp_hash": sp_hash.hex()})
                return {
                    "signage_point": SignagePoint.from_json_dict(response["signage_point"]),
                    "time_received": response["time_received"],
                    "reverted": response["reverted"],
                }
            else:
                assert challenge_hash is not None
                response = await self.fetch("get_recent_signage_point_or_eos", {"challenge_hash": challenge_hash.hex()})
                return {
                    "eos": EndOfSubSlotBundle.from_json_dict(response["eos"]),
                    "time_received": response["time_received"],
                    "reverted": response["reverted"],
                }
        except Exception:
            return None

    async def get_fee_estimate(
        self,
        target_times: Optional[list[int]],
        cost: Optional[int],
    ) -> dict[str, Any]:
        response = await self.fetch("get_fee_estimate", {"cost": cost, "target_times": target_times})
        return response



from __future__ import annotations

import logging
from typing import Any

from chik_rs import ConsensusConstants as ConsensusConstants  # noqa: PLC0414

from chik.util.byte_types import hexstr_to_bytes
from chik.util.hash import std_hash

log = logging.getLogger(__name__)


def replace_str_to_bytes(constants: ConsensusConstants, **changes: Any) -> ConsensusConstants:
    """
    Overrides str (hex) values with bytes.
    """

    filtered_changes = {}
    for k, v in changes.items():
        if not hasattr(constants, k):
            # NETWORK_TYPE used to be present in default config, but has been removed
            if k not in {"NETWORK_TYPE"}:
                log.warning(f'invalid key in network configuration (config.yaml) "{k}". Ignoring')
            continue
        if isinstance(v, str):
            filtered_changes[k] = hexstr_to_bytes(v)
        else:
            filtered_changes[k] = v

    # if we override the additional data (replay protection across forks)
    # make sure the other variants of the AGG_SIG_* conditions are also covered
    if "AGG_SIG_ME_ADDITIONAL_DATA" in filtered_changes:
        AGG_SIG_DATA = filtered_changes["AGG_SIG_ME_ADDITIONAL_DATA"]
        if "AGG_SIG_PARENT_ADDITIONAL_DATA" not in filtered_changes:
            filtered_changes["AGG_SIG_PARENT_ADDITIONAL_DATA"] = std_hash(AGG_SIG_DATA + bytes([43]))
        if "AGG_SIG_PUZZLE_ADDITIONAL_DATA" not in filtered_changes:
            filtered_changes["AGG_SIG_PUZZLE_ADDITIONAL_DATA"] = std_hash(AGG_SIG_DATA + bytes([44]))
        if "AGG_SIG_AMOUNT_ADDITIONAL_DATA" not in filtered_changes:
            filtered_changes["AGG_SIG_AMOUNT_ADDITIONAL_DATA"] = std_hash(AGG_SIG_DATA + bytes([45]))
        if "AGG_SIG_PUZZLE_AMOUNT_ADDITIONAL_DATA" not in filtered_changes:
            filtered_changes["AGG_SIG_PUZZLE_AMOUNT_ADDITIONAL_DATA"] = std_hash(AGG_SIG_DATA + bytes([46]))
        if "AGG_SIG_PARENT_AMOUNT_ADDITIONAL_DATA" not in filtered_changes:
            filtered_changes["AGG_SIG_PARENT_AMOUNT_ADDITIONAL_DATA"] = std_hash(AGG_SIG_DATA + bytes([47]))
        if "AGG_SIG_PARENT_PUZZLE_ADDITIONAL_DATA" not in filtered_changes:
            filtered_changes["AGG_SIG_PARENT_PUZZLE_ADDITIONAL_DATA"] = std_hash(AGG_SIG_DATA + bytes([48]))

    # TODO: this is too magical here and is really only used for configuration unmarshalling
    return constants.replace(**filtered_changes)  # type: ignore[arg-type]



from __future__ import annotations

import click

from chik.cmds.cmd_classes import ChikCliContext
from chik.plotters.plotters import call_plotters


@click.command(
    "plotters",
    help="Advanced plotting options",
    context_settings={"ignore_unknown_options": True},
    add_help_option=False,
)
@click.pass_context
@click.argument("args", nargs=-1)
def plotters_cmd(ctx: click.Context, args: tuple[click.Argument]) -> None:
    call_plotters(ChikCliContext.set_default(ctx).root_path, args)



from __future__ import annotations

from io import TextIOWrapper
from typing import Optional

import click

from chik import __version__
from chik.cmds.beta import beta_cmd
from chik.cmds.cmd_classes import ChikCliContext
from chik.cmds.completion import completion
from chik.cmds.configure import configure_cmd
from chik.cmds.dao import dao_cmd
from chik.cmds.data import data_cmd
from chik.cmds.db import db_cmd
from chik.cmds.dev import dev_cmd
from chik.cmds.farm import farm_cmd
from chik.cmds.init import init_cmd
from chik.cmds.keys import keys_cmd
from chik.cmds.netspace import netspace_cmd
from chik.cmds.passphrase import passphrase_cmd
from chik.cmds.peer import peer_cmd
from chik.cmds.plotnft import plotnft_cmd
from chik.cmds.plots import plots_cmd
from chik.cmds.plotters import plotters_cmd
from chik.cmds.rpc import rpc_cmd
from chik.cmds.show import show_cmd
from chik.cmds.start import start_cmd
from chik.cmds.stop import stop_cmd
from chik.cmds.wallet import wallet_cmd
from chik.util.default_root import DEFAULT_KEYS_ROOT_PATH, resolve_root_path
from chik.util.errors import KeychainCurrentPassphraseIsInvalid
from chik.util.keychain import Keychain, set_keys_root_path
from chik.util.ssl_check import check_ssl

CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
    "show_default": True,
}


@click.group(
    help=f"\n  Manage chik blockchain infrastructure ({__version__})\n",
    epilog="Try 'chik start node', 'chik netspace -d 192', or 'chik show -s'",
    context_settings=CONTEXT_SETTINGS,
)
@click.option(
    "--root-path",
    default=resolve_root_path(override=None),
    help="Config file root",
    type=click.Path(),
    show_default=True,
)
@click.option(
    "--keys-root-path", default=DEFAULT_KEYS_ROOT_PATH, help="Keyring file root", type=click.Path(), show_default=True
)
@click.option("--passphrase-file", type=click.File("r"), help="File or descriptor to read the keyring passphrase from")
@click.pass_context
def cli(
    ctx: click.Context,
    root_path: str,
    keys_root_path: Optional[str] = None,
    passphrase_file: Optional[TextIOWrapper] = None,
) -> None:
    from pathlib import Path

    context = ChikCliContext.set_default(ctx=ctx)
    context.root_path = Path(root_path)

    # keys_root_path and passphrase_file will be None if the passphrase options have been
    # scrubbed from the CLI options
    if keys_root_path is not None:
        set_keys_root_path(Path(keys_root_path))

    if passphrase_file is not None:
        import sys

        from chik.cmds.passphrase_funcs import cache_passphrase, read_passphrase_from_file

        try:
            passphrase = read_passphrase_from_file(passphrase_file)
            if Keychain.master_passphrase_is_valid(passphrase):
                cache_passphrase(passphrase)
            else:
                raise KeychainCurrentPassphraseIsInvalid()
        except KeychainCurrentPassphraseIsInvalid:
            if Path(passphrase_file.name).is_file():
                print(f'Invalid passphrase found in "{passphrase_file.name}"')
            else:
                print("Invalid passphrase")
            sys.exit(1)
        except Exception as e:
            print(f"Failed to read passphrase: {e}")

    check_ssl(Path(root_path))


@cli.command("version", help="Show chik version")
def version_cmd() -> None:
    print(__version__)


@cli.command("run_daemon", help="Runs chik daemon")
@click.option(
    "--wait-for-unlock",
    help="If the keyring is passphrase-protected, the daemon will wait for an unlock command before accessing keys",
    default=False,
    is_flag=True,
    hidden=True,  # --wait-for-unlock is only set when launched by chik start <service>
)
@click.pass_context
def run_daemon_cmd(ctx: click.Context, wait_for_unlock: bool) -> None:
    import asyncio

    from chik.daemon.server import async_run_daemon
    from chik.util.keychain import Keychain

    wait_for_unlock = wait_for_unlock and Keychain.is_keyring_locked()

    asyncio.run(async_run_daemon(ChikCliContext.set_default(ctx).root_path, wait_for_unlock=wait_for_unlock))


cli.add_command(keys_cmd)
cli.add_command(plots_cmd)
cli.add_command(wallet_cmd)
cli.add_command(plotnft_cmd)
cli.add_command(configure_cmd)
cli.add_command(init_cmd)
cli.add_command(rpc_cmd)
cli.add_command(show_cmd)
cli.add_command(start_cmd)
cli.add_command(stop_cmd)
cli.add_command(netspace_cmd)
cli.add_command(farm_cmd)
cli.add_command(plotters_cmd)
cli.add_command(db_cmd)
cli.add_command(peer_cmd)
cli.add_command(data_cmd)
cli.add_command(passphrase_cmd)
cli.add_command(beta_cmd)
cli.add_command(completion)
cli.add_command(dao_cmd)
cli.add_command(dev_cmd)


def main() -> None:
    cli()


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Callable, Optional

from chik.consensus.blockchain import StateChangeSummary
from chik.types.blockchain_format.sized_bytes import bytes32


def get_hints_and_subscription_coin_ids(
    state_change_summary: StateChangeSummary,
    has_coin_subscription: Callable[[bytes32], bool],
    has_puzzle_subscription: Callable[[bytes32], bool],
) -> tuple[list[tuple[bytes32, bytes]], list[bytes32]]:
    # Precondition: all hints passed in are max 32 bytes long
    # Returns the hints that we need to add to the DB, and the coin ids that need to be looked up

    # Finds the coin IDs that we need to lookup in order to notify wallets of hinted transactions
    hint: Optional[bytes]
    hints_to_add: list[tuple[bytes32, bytes]] = []

    # Goes through additions and removals for each block and flattens to a map and a set
    lookup_coin_ids: set[bytes32] = set()

    def add_if_coin_subscription(coin_id: bytes32) -> None:
        if has_coin_subscription(coin_id):
            lookup_coin_ids.add(coin_id)

    def add_if_ph_subscription(puzzle_hash: bytes32, coin_id: bytes32) -> None:
        if has_puzzle_subscription(puzzle_hash):
            lookup_coin_ids.add(coin_id)

    for spend_id, puzzle_hash in state_change_summary.removals:
        # Record all coin_ids that we are interested in, that had changes
        add_if_coin_subscription(spend_id)
        add_if_ph_subscription(puzzle_hash, spend_id)

    for addition_coin, hint in state_change_summary.additions:
        addition_coin_name = addition_coin.name()
        add_if_coin_subscription(addition_coin_name)
        add_if_ph_subscription(addition_coin.puzzle_hash, addition_coin_name)
        if hint is None:
            continue
        if len(hint) == 32:
            add_if_ph_subscription(bytes32(hint), addition_coin_name)

        if len(hint) > 0:
            assert len(hint) <= 32
            hints_to_add.append((addition_coin_name, hint))

    # Goes through all new reward coins
    for reward_coin in state_change_summary.new_rewards:
        reward_coin_name: bytes32 = reward_coin.name()
        add_if_coin_subscription(reward_coin_name)
        add_if_ph_subscription(reward_coin.puzzle_hash, reward_coin_name)

    return hints_to_add, list(lookup_coin_ids)
